# coding: utf-8
"""
Scoring utilities
"""
import numpy as np
import pandas as pd

from .pretty_print import IS_COLOR_TERM
from .pretty_print import print_warning


def reorder_df_scores(df_scores, score_types):
    """Reorder scores according to the order in score_types.

    Parameters
    ----------
    df_scores : pd.DataFrame
        the score dataframe
    score_types : list of score types

    Returns
    -------
    df_scores : the dataframe with reordered scores
    """
    try:
        # try to re-order columns/rows in the printed array
        # we may not have all train, valid, test, so need to select
        index_order = np.array(['train', 'valid', 'test'])
        ordered_index = index_order[np.isin(index_order, df_scores.index)]
        df_scores = df_scores.loc[
            ordered_index, [score_type.name for score_type in score_types]]
    except Exception:
        print_warning("Couldn't re-order the score matrix..")
    return df_scores


def mean_score_matrix(df_scores_list, score_types):
    u"""Construct a mean ± std score dataframe from a list of score dataframes.

    Parameters
    ----------
    df_scores_list : list of pd.DataFrame
        a list of score data frames to average
    score_types : list of score types
        a list of score types to use

    Returns
    -------
    df_scores : the mean ± std score dataframe

    """
    scores = np.array([df_scores.values for df_scores in df_scores_list])
    meanss = scores.mean(axis=0)
    stdss = scores.std(axis=0)
    precisions = [st.precision for st in score_types]
    precisions.append(1)  # for time
    # we use unicode no break space so split in print_df_scores works
    if IS_COLOR_TERM:
        strs = np.array([[
            u'{val}\u00A0±\u00A0{std}'.format(
                val=round(mean, prec),
                std=round(std, prec + 1))
            for mean, std, prec in zip(means, stds, precisions)]
            for means, stds in zip(meanss, stdss)])
    else:
        strs = np.array([[
            u'{val} +- {std}'.format(
                val=round(mean, prec),
                std=round(std, prec + 1))
            for mean, std, prec in zip(means, stds, precisions)]
            for means, stds in zip(meanss, stdss)])
    df_scores = pd.DataFrame(
        strs, columns=df_scores_list[0].columns, index=df_scores_list[0].index)
    return df_scores


def score_matrix_from_scores(score_types, steps, scoress):
    """Construct a score dataframe from a matrix of scores.

    Parameters
    ----------
    score_types : list of score types
        a list of score types to use, score_type.name serves as column index
    steps : a list of strings
        subset of ['train', 'valid', 'test'], serves as row index

    Returns
    -------
    df_scores : the score dataframe
    """
    results = []
    for step, scores in zip(steps, scoress):
        for score_type, score in zip(score_types, scores):
            results.append(
                {'step': str(step), 'score': score_type.name, 'value': score})
    df_scores = pd.DataFrame(results)
    df_scores = df_scores.set_index(['step', 'score'])['value']
    df_scores = df_scores.unstack()
    df_scores = reorder_df_scores(df_scores, score_types)
    return df_scores


def score_matrix(score_types, ground_truth, predictions):
    """Construct a score dataframe by scoring predictions against ground truth.

    Parameters
    ----------
    score_types : list of score types
        a list of score types to use, score_type.name serves as column index
    ground_truth : dict of Predictions
        the ground truth data
    predictions : dict of Predictions
        the predicted data

    Returns
    -------
    df_scores : pd.DataFrame
        table of scores (rows = train/valid/test steps, columns = scores)
    """
    if set(ground_truth.keys()) != set(predictions.keys()):
        raise ValueError(('Predictions and ground truth steps '
                          'do not match:\n'
                          ' * predictions = {} \n'
                          ' * ground_truth = {} ')
                         .format(set(predictions.keys()),
                                 set(ground_truth.keys())))
    steps = ground_truth.keys()
    scoress = [[
        score_type.score_function(ground_truth[step], predictions[step])
        for score_type in score_types] for step in ground_truth]
    return score_matrix_from_scores(score_types, steps, scoress)


def round_df_scores(df_scores, score_types):
    """Round scores to the precision set in the score type.

    Parameters
    ----------
    df_scores : pd.DataFrame
        the score dataframe
    score_types : list of score types

    Returns
    -------
    df_scores : the dataframe with rounded scores
    """
    df_scores_copy = df_scores.copy()
    for column, score_type in zip(df_scores_copy, score_types):
        df_scores_copy[column] = [round(score, score_type.precision)
                                  for score in df_scores_copy[column]]
    return df_scores_copy



"""A generalization of the classification accuracy with cross-class scores.

Soften the accuracy score by giving scores through certain misclassifications
defined by the score matrix. For example, in ordinal regression we may want
not to penalize too much misclassifications to neighbor classes. The score also
generalizes RMSE-like regression scores for ordinal regression (when true and
predicted output levels are coming from a fixed set) by allowing to define
arbitrary misclassification scores.
"""

from __future__ import division

import numpy as np
from .base import BaseScoreType


class SoftAccuracy(BaseScoreType):
    is_lower_the_better = False
    minimum = 0.0

    def __init__(self, score_matrix, name='soft precision', precision=2):
        self.name = name
        self.precision = precision
        self.maximum = np.max(score_matrix)
        self.score_matrix = score_matrix

    def __call__(self, y_true_proba, y_proba):
        # Clip negative probas
        y_proba_positive = np.clip(y_proba, 0, 1)
        # normalize rows
        y_proba_normalized = y_proba_positive / np.sum(
            y_proba_positive, axis=1, keepdims=True)
        # Smooth true probabilities with score_matrix
        y_true_smoothed = y_true_proba.dot(self.score_matrix)
        # Compute dot product between the predicted probabilities and
        # the smoothed true "probabilites" ("" because it does not sum to 1)
        scores = np.sum(y_proba_normalized * y_true_smoothed, axis=1)
        scores = np.nan_to_num(scores)
        score = np.mean(scores)
        # to pick up all zero probabilities
        score = np.nan_to_num(score)
        return score



from __future__ import division

import os

import numpy as np

from ..utils.importing import import_module_from_source


class ObjectDetector(object):
    """
    Object detection workflow.

    This workflow is used to train image object detection tasks, typically
    when the dataset cannot be stored in memory.
    Submissions need to contain two files, which by default are named:
    image_preprocessor.py and object_detector_model.py (they can be
    modified by changing `workflow_element_names`).

    image_preprocessor.py needs a `transform` function, which
    is used for preprocessing the images. It takes an image as input
    and it returns an image as an output. Optionally, image_preprocessor.py
    can also have a function `transform_test`, which is used only to preprocess
    images at test time. Otherwise, if `transform_test` does not exist,
    `transform` is used at train and test time.

    object_detector_model.py needs a `ObjectDetector` class, which
    implements `fit` and `predict`.

    Parameters
    ==========

    test_batch_size : int
        batch size used for testing.

    chunk_size : int
        size of the chunk used to load data from disk into memory.
        (see at the top of the file what a chunk is and its difference
         with the mini-batch size of neural nets).

    n_jobs : int
        the number of jobs used to load images from disk to memory as `chunks`.

    """

    def __init__(self, workflow_element_names=['object_detector']):
        self.element_names = workflow_element_names

    def train_submission(self, module_path, X, y, train_is=None):
        """Train a ObjectDetector.

        module_path : str
            module where the submission is. the folder of the module
            have to contain object_detector.py.
        X : ArrayContainer vector of int
            vector of image data to train on
        y : vector of lists
            vector of object labels corresponding to X
        train_is : vector of int
           indices from X_array to train on
        """
        if train_is is None:
            train_is = slice(None, None, None)

        # object detector model
        detector = import_module_from_source(
            os.path.join(module_path, self.element_names[0] + '.py'),
            self.element_names[0],
            sanitize=True
        )
        clf = detector.ObjectDetector()

        # train and return fitted model
        clf.fit(X[train_is], y[train_is])
        return clf

    def test_submission(self, trained_model, X):
        """Test an ObjectDetector.

        trained_model
            Trained model returned by `train_submission`.
        X : ArrayContainer of int
            Vector of image data to test on.
        """
        clf = trained_model
        y_pred = clf.predict(X)
        return y_pred


class BatchGeneratorBuilder(object):
    """A batch generator builder for generating batches of images on the fly.

    This class is a way to build training and
    validation generators that yield each time a tuple (X, y) of mini-batches.
    The generators are built in a way to fit into keras API of `fit_generator`
    (see https://keras.io/models/model/).

    The fit function from `Classifier` should then use the instance
    to build train and validation generators, using the method
    `get_train_valid_generators`

    Parameters
    ==========

    X_array : ArrayContainer of int
        vector of image data to train on
    y_array : vector of int
        vector of object labels corresponding to `X_array`

    """

    def __init__(self, X_array, y_array):
        self.X_array = X_array
        self.y_array = y_array
        self.nb_examples = len(X_array)

    def get_train_valid_generators(self, batch_size=256, valid_ratio=0.1):
        """Build train and valid generators for keras.

        This method is used by the user defined `Classifier` to o build train
        and valid generators that will be used in keras `fit_generator`.

        Parameters
        ==========

        batch_size : int
            size of mini-batches
        valid_ratio : float between 0 and 1
            ratio of validation data

        Returns
        =======

        a 4-tuple (gen_train, gen_valid, nb_train, nb_valid) where:
            - gen_train is a generator function for training data
            - gen_valid is a generator function for valid data
            - nb_train is the number of training examples
            - nb_valid is the number of validation examples
        The number of training and validation data are necessary
        so that we can use the keras method `fit_generator`.
        """
        nb_valid = int(valid_ratio * self.nb_examples)
        nb_train = self.nb_examples - nb_valid
        indices = np.arange(self.nb_examples)
        train_indices = indices[0:nb_train]
        valid_indices = indices[nb_train:]
        gen_train = self._get_generator(
            indices=train_indices, batch_size=batch_size)
        gen_valid = self._get_generator(
            indices=valid_indices, batch_size=batch_size)
        return gen_train, gen_valid, nb_train, nb_valid

    def _get_generator(self, indices=None, batch_size=256):
        if indices is None:
            indices = np.arange(self.nb_examples)
        # Infinite loop, as required by keras `fit_generator`.
        # However, as we provide the number of examples per epoch
        # and the user specifies the total number of epochs, it will
        # be able to end.
        while True:
            X = self.X_array[indices]
            y = self.y_array[indices]

            # converting to float needed?
            # X = np.array(X, dtype='float32')

            # Yielding mini-batches
            for i in range(0, len(X), batch_size):
                yield X[i:i + batch_size], y[i:i + batch_size]



"""A supervised clustering or transfer learning workflow.

Training and test data consist of a set of events ad each event has an
unknown number of clusters. At training time we know the classes (clusters) of
each event. At test time we only know which instances belong to the same event.
Events share statistical properties which can be exploited.

From another point of view, the setup is transfer learning: each event is
a new task, and classed in the task share statistical properties.

User submissions implement a fit function that receives both event ids and
cluster ids (besides input covariates). At test time `test_submission`
receives only the coavariates and the event_id (assumed to be the first column
of `X_array`). It slices up `X_array` into single events, drops the event ids,
and sends the single event to the `predict_single_event` function implemented
by the users. This function returns a vector of labels (cluster assignments)
which is then joined back the the event id column and returned (to be passed
into `prediction_types.Clustering` and evaluated by
`score_types.clustering_efficiency`).
"""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause

import os

import numpy as np

from ..utils.importing import import_module_from_source


class Clusterer(object):
    def __init__(self, workflow_element_names=['clusterer']):
        self.element_names = workflow_element_names

    def train_submission(self, module_path, X_array, y_array, train_is=None):
        if train_is is None:
            train_is = slice(None, None, None)
        clusterer = import_module_from_source(
            os.path.join(module_path, self.element_names[0] + '.py'),
            self.element_names[0],
            sanitize=True
        )
        ctr = clusterer.Clusterer()
        ctr.fit(X_array[train_is], y_array[train_is])
        return ctr

    def test_submission(self, trained_model, X_array):
        ctr = trained_model
        unique_event_ids = np.unique(X_array[:, 0])
        cluster_ids = np.empty(len(X_array), dtype='int')

        for event_id in unique_event_ids:
            event_indices = (X_array[:, 0] == event_id)
            # select an event and drop event ids
            X_event = X_array[event_indices][:, 1:]
            cluster_ids[event_indices] = ctr.predict_single_event(X_event)

        return np.stack(
            (X_array[:, 0], cluster_ids), axis=-1).astype(dtype='int')



"""Regression predictions.

y_pred can be one or two-dimensional (for multi-target regression)
"""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause

import numpy as np
from .base import BasePrediction


def _regression_init(self, y_pred=None, y_true=None, n_samples=None):
    if y_pred is not None:
        self.y_pred = np.array(y_pred)
    elif y_true is not None:
        self.y_pred = np.array(y_true)
    elif n_samples is not None:
        if self.n_columns == 0:
            shape = (n_samples)
        else:
            shape = (n_samples, self.n_columns)
        self.y_pred = np.empty(shape, dtype=float)
        self.y_pred.fill(np.nan)
    else:
        raise ValueError(
            'Missing init argument: y_pred, y_true, or n_samples')
    self.check_y_pred_dimensions()


def make_regression(label_names=[]):
    Predictions = type(
        'Regression',
        (BasePrediction,),
        {'label_names': label_names,
         'n_columns': len(label_names),
         'n_columns_true': len(label_names),
         '__init__': _regression_init,
         })
    return Predictions



from .base import BaseScoreType


class Combined(BaseScoreType):
    is_lower_the_better = None
    minimum = 0.0
    maximum = 0.0

    def __init__(self, score_types, weights, name='combined', precision=2):
        self.name = name
        self.score_types = score_types
        self.weights = weights
        self.precision = precision
        for weight, score_type in zip(self.weights, self.score_types):
            if self.is_lower_the_better is None:
                self.is_lower_the_better = score_type.is_lower_the_better
            elif self.is_lower_the_better != score_type.is_lower_the_better:
                raise ValueError(
                    'Cannot combine scores of lower and higher the better')
            self.minimum += weight * score_type.minimum
            self.maximum += weight * score_type.maximum

    def score_function(self, ground_truths_combined, predictions_combined,
                       valid_indexes=None):
        score = 0.0
        for weight, score_type, ground_truths, predictions in zip(
                self.weights, self.score_types,
                ground_truths_combined.predictions_list,
                predictions_combined.predictions_list):
            score += weight * score_type.score_function(
                ground_truths, predictions, valid_indexes)
        return score

    def __call__(self, y_true, y_pred):
        raise ValueError('Combined score has no deep score function.')



from __future__ import division

import numpy as np

from .base import BaseScoreType
from .util import _filter_y_pred
from .iou import cc_iou
from .precision_recall import precision, recall


class DetectionAveragePrecision(BaseScoreType):
    is_lower_the_better = False
    minimum = 0.0
    maximum = 1.0
    _name = 'average_precision'

    def __init__(self, name=_name, precision=3, iou_threshold=0.5):
        if name is None:
            self.name = '{name}(IOU={iou_threshold})'.format(
                name=self._name, iou_threshold=iou_threshold)
        else:
            self.name = name
        self.precision = precision
        self.iou_threshold = iou_threshold

    def __call__(self, y_true, y_pred):
        _, ps, rs = precision_recall_curve_greedy(
            y_true, y_pred, iou_threshold=self.iou_threshold)
        return average_precision_exact(ps, rs)


AverageDetectionPrecision = DetectionAveragePrecision


def precision_recall_curve(y_true, y_pred, conf_thresholds, iou_threshold=0.5):
    """
    Calculate precision and recall for different confidence thresholds.

    Parameters
    ----------
    y_true : list of list of tuples
        Tuples are of form (x, y, radius).
    y_pred : list of list of tuples
        Tuples are of form (x, y, radius, confidence).
    conf_thresholds : array-like
        The confidence threshold for which to calculate the
        precision and recall.
    iou_threshold : float
        Threshold to determine match.

    Returns
    -------
    ps, rs : arrays with precisions, recalls for each confidence threshold

    """
    ps = []
    rs = []

    for conf_threshold in conf_thresholds:
        y_pred_above_confidence = _filter_y_pred(y_pred, conf_threshold)
        ps.append(precision(
            y_true, y_pred_above_confidence, iou_threshold=iou_threshold))
        rs.append(recall(
            y_true, y_pred_above_confidence, iou_threshold=iou_threshold))

    return np.array(ps), np.array(rs)


def _add_id(y):
    """
    Helper function to flatten and add id column to list of lists.

    Since the list of lists is flattened into a single array, empty lists
    do not result in an entry in the final array.

    """
    y_new = []

    for i, y_patch in enumerate(y):
        if len(y_patch):
            tmp = np.asarray(y_patch)
            tmp = np.insert(tmp, 0, i, axis=1)
            y_new.append(tmp)

    if y_new:
        return np.vstack(y_new)
    else:
        return np.array([[]])


def precision_recall_curve_greedy(y_true, y_pred, iou_threshold=0.5):
    """
    Calculate precision and recall incrementally based on the predictions
    sorted by confidence (not calculating an exact optimal match for each
    confidence level).

    Parameters
    ----------
    y_true : list of lists of (x, y, r) tuples
    y_pred : list of lists of (conf, x, y, r) tuples
    iou_threshold : float [0 - 1], default 0.5

    Returns
    -------
    Three arrays:

    confidence_values
        The flattened and sorted confidence values of y_pred
    precision
        The corresponding precision values
    recall
        The corresponding recall values

    """
    # flatten y_pred into single array and add column with img id
    y_pred2 = _add_id(y_pred)
    if not y_pred2.size:
        return np.array([]), np.array([]), np.array([])

    # Sorting predicted objects by decreasing confidence
    y_pred2_sorted = y_pred2[np.argsort(y_pred2[:, 1])[::-1], :]

    # array to store whether a match is observed or not for each prediction
    res = np.empty(len(y_pred2_sorted), dtype='bool')

    # object to keep track of matches: (img id, object index) pairs
    matched = set([])
    confs = []

    for i, pred in enumerate(y_pred2_sorted):
        patch_id, conf, row, col, rad = pred

        y_patch = y_true[int(patch_id)]
        n_true = len(y_patch)

        if n_true > 0:
            ious = np.empty(n_true)

            for j in range(n_true):
                ious[j] = cc_iou(y_patch[j], (row, col, rad))

            i_max = np.argmax(ious)
            if ((ious[i_max] > iou_threshold)
                    and (patch_id, i_max) not in matched):
                res[i] = True
                # add match identifier to set to later check
                # we don't have a duplicate match
                matched.add((patch_id, i_max))
            else:
                res[i] = False
        else:
            res[i] = False

        confs.append(conf)

    n_true_total = np.sum([len(x) for x in y_true])

    recall = np.cumsum(res) / n_true_total
    precision = np.cumsum(res) / np.arange(1, len(res) + 1)

    return np.array(confs), precision, recall


def average_precision_interpolated(ps, rs):
    """
    The Average Precision (AP) score.

    Calculation based on the 11-point interpolation of the precision-recall
    curve (method described for Pascal VOC challenge,
    http://homepages.inf.ed.ac.uk/ckiw/postscript/ijcv_voc09.pdf).

    TODO: they changed this in later:
    http://homepages.inf.ed.ac.uk/ckiw/postscript/ijcv_voc09.pdf

    https://stackoverflow.com/questions/36274638/map-metric-in-object-detection-and-computer-vision  # noqa

    Parameters
    ----------
    ps, rs : arrays of same length with corresponding precision / recall scores

    Returns
    -------
    ap : int [0 - 1]
        Average precision score

    """
    ps = np.asarray(ps)
    rs = np.asarray(rs)

    p_at_r = []

    for r in np.arange(0, 1.1, 0.1):
        p = np.array(ps)[np.array(rs) >= r]
        if p.size:
            p_at_r.append(np.nanmax(p))
        else:
            p_at_r.append(0)

    ap = np.mean(p_at_r)
    return ap


def average_precision_exact(ps, rs):
    # from https://github.com/amdegroot/ssd.pytorch/blob/ce4c994db0ee11f82aabb4fdb3499dc970156db5/eval.py#L182-L213  # noqa

    # correct AP calculation
    # first append sentinel values at the end
    mrec = np.concatenate(([0.], rs, [1.]))
    mpre = np.concatenate(([0.], ps, [0.]))

    # compute the precision envelope
    for i in range(mpre.size - 1, 0, -1):
        mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

    # to calculate area under PR curve, look for points
    # where X axis (recall) changes value
    i = np.where(mrec[1:] != mrec[:-1])[0]

    # and sum (\Delta recall) * prec
    ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])

    return ap



from .accuracy import Accuracy
from .balanced_accuracy import BalancedAccuracy
from .base import BaseScoreType
from .brier_score import (
    BrierScore, BrierSkillScore, BrierScoreReliability, BrierScoreResolution)
from .clustering_efficiency import ClusteringEfficiency
from .classification_error import ClassificationError
from .combined import Combined
from .detection import (
    OSPA, SCP, DetectionPrecision, DetectionRecall, MADCenter, MADRadius,
    AverageDetectionPrecision, DetectionAveragePrecision)
from .f1_above import F1Above
from .macro_averaged_recall import MacroAveragedRecall
from .make_combined import MakeCombined
from .mare import MARE
from .negative_log_likelihood import NegativeLogLikelihood
from .normalized_gini import NormalizedGini
from .normalized_rmse import NormalizedRMSE
from .relative_rmse import RelativeRMSE
from .rmse import RMSE
from .roc_auc import ROCAUC
from .soft_accuracy import SoftAccuracy

__all__ = [
    'Accuracy',
    'BalancedAccuracy',
    'BaseScoreType',
    'BrierScore',
    'BrierScoreReliability',
    'BrierScoreResolution',
    'BrierSkillScore',
    'ClassificationError',
    'ClusteringEfficiency',
    'Combined',
    'DetectionPrecision',
    'DetectionRecall',
    'DetectionAveragePrecision',
    'F1Above',
    'MacroAveragedRecall',
    'MakeCombined',
    'MADCenter',
    'MADRadius',
    'MARE',
    'NegativeLogLikelihood',
    'NormalizedGini',
    'NormalizedRMSE',
    'OSPA',
    'RelativeRMSE',
    'RMSE',
    'ROCAUC',
    'SCP',
    'SoftAccuracy',
]



from .clustering import make_clustering
from .combined import make_combined
from .detection import make_detection
from .multiclass import make_multiclass
from .regression import make_regression

__all__ = [
    'make_clustering',
    'make_combined',
    'make_detection',
    'make_multiclass',
    'make_regression',
]



"""Clustering predictions.

Predictions have two columns: event_id and cluster_id. event_id
comes from the first column of X when test is called. See
``rampwf.workflows.clustering``.

The predictions are used in "supervised" clustering where the data
consists of a set of events ad each event has an unknown number of
clusters. At training time we know the classes (clusters) of each event.
At test time we only know which instances belong to the same event. Events
share statistical properties which can be exploited.

From another point of view, the setup is transfer learning: each event is
a new task, and classed in the task share statistical properties.
"""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause

import numpy as np
from .base import BasePrediction


def _clustering_init(self, y_pred=None, y_true=None, n_samples=None):
    if y_pred is not None:
        self.y_pred = np.array(y_pred)
    elif y_true is not None:
        self.y_pred = np.array(y_true)
    elif n_samples is not None:
        self.y_pred = np.empty((n_samples, 2), dtype=float)
        self.y_pred.fill(np.nan)
    else:
        raise ValueError(
            'Missing init argument: y_pred, y_true, or n_samples')
    self.check_y_pred_dimensions()


@property
def _valid_indexes(self):
    """Return valid indices (e.g., a cross-validation slice)."""
    return ~np.isnan(self.y_pred[:, 1])

    return ~np.isnan(self.y_pred[:, 0])


def make_clustering():
    Predictions = type(
        'Clustering',
        (BasePrediction,),
        {'n_columns': 2,
         'n_columns_true': 2,
         '__init__': _clustering_init,
         'valid_indexes': _valid_indexes,
         })
    return Predictions



from .air_passengers import AirPassengers
from .classifier import Classifier
from .clusterer import Clusterer
from .drug_spectra import DrugSpectra
from .el_nino import ElNino
from .feature_extractor import FeatureExtractor
from .feature_extractor_classifier import FeatureExtractorClassifier
from .feature_extractor_regressor import FeatureExtractorRegressor
from .image_classifier import ImageClassifier
from .simplified_image_classifier import SimplifiedImageClassifier
from .object_detector import ObjectDetector
from .regressor import Regressor
from .ts_feature_extractor import TimeSeriesFeatureExtractor
from .grid_feature_extractor_classifier import GridFeatureExtractorClassifier
from .sklearn_pipeline import SKLearnPipeline, Estimator, EstimatorExternalData

__all__ = [
    'AirPassengers',
    'Classifier',
    'Clusterer',
    'DrugSpectra',
    'ElNino',
    'FeatureExtractor',
    'FeatureExtractorClassifier',
    'FeatureExtractorRegressor',
    'ImageClassifier',
    'SimplifiedImageClassifier',
    'ObjectDetector',
    'Regressor',
    'TimeSeriesFeatureExtractor',
    'GridFeatureExtractorClassifier',
    'SKLearnPipeline',
    'Estimator',
    'EstimatorExternalData'
]



from .command_line import (
    ramp_test_submission, ramp_test_notebook, ramp_convert_notebook,
    ramp_blend_submissions)
from .testing import (
    assert_title, assert_data, assert_cv, assert_read_problem,
    assert_submission, assert_notebook, blend_submissions)
from .submission import run_submission_on_cv_fold
from .combine import get_score_cv_bags
from .importing import import_module_from_source

__all__ = ['assert_cv',
           'assert_data',
           'assert_notebook',
           'assert_read_problem',
           'assert_submission',
           'assert_title',
           'blend_submissions',
           'get_score_cv_bags',
           'import_module_from_source',
           'ramp_blend_submissions',
           'ramp_convert_notebook',
           'ramp_test_notebook',
           'ramp_test_submission',
           'run_submission_on_cv_fold',
           ]



from __future__ import division

from ..base import BaseScoreType
from .util import _filter_y_pred


class DetectionBaseScoreType(BaseScoreType):
    """Common abstract base type for detection scores.

    Implements `__call__` by selecting all prediction detections with
    a confidence higher than `conf_threshold`. It assumes that the child
    class implements `detection_score`.
    """

    conf_threshold = 0.5

    def __call__(self, y_true, y_pred, conf_threshold=None):
        if conf_threshold is None:
            conf_threshold = self.conf_threshold
        y_pred_above_confidence = _filter_y_pred(y_pred, conf_threshold)

        return self.detection_score(y_true, y_pred_above_confidence)


class DetectionBaseIOUScoreType(DetectionBaseScoreType):

    def __init__(self, name=None, precision=3, conf_threshold=0.5,
                 minipatch=None, iou_threshold=0.5):
        if name is None:
            self.name = '{name}(IOU={iou_threshold})'.format(
                name=self._name, iou_threshold=iou_threshold)
        else:
            self.name = name
        self.precision = precision
        self.conf_threshold = conf_threshold
        self.minipatch = minipatch
        self.iou_threshold = iou_threshold



import numpy as np
from sklearn.metrics import log_loss

from .base import BaseScoreType


class NegativeLogLikelihood(BaseScoreType):
    is_lower_the_better = True
    minimum = 0.0
    maximum = float('inf')

    def __init__(self, name='negative log likelihood', precision=2):
        self.name = name
        self.precision = precision

    def __call__(self, y_true_proba, y_proba):
        score = log_loss(y_true_proba, y_proba)
        return score



from sklearn.metrics import recall_score
from .classifier_base import ClassifierBaseScoreType


class MacroAveragedRecall(ClassifierBaseScoreType):
    is_lower_the_better = False
    minimum = 0.0
    maximum = 1.0

    def __init__(self, name='accuracy', precision=2):
        self.name = name
        self.precision = precision

    def __call__(self, y_true_label_index, y_pred_label_index):
        score = recall_score(
            y_true_label_index, y_pred_label_index, average='macro')
        return score



"""Hyperparameter optiomization for ramp-kits."""
from __future__ import print_function

import re
import os
import shutil
import numpy as np
import pandas as pd
from tempfile import mkdtemp
from ..utils import (
    assert_read_problem, import_module_from_source, run_submission_on_cv_fold)

HYPERPARAMS_SECTION_START = '# RAMP START HYPERPARAMETERS'
HYPERPARAMS_SECTION_END = '# RAMP END HYPERPARAMETERS'
HYPERPARAMS_REPL_REGEX = re.compile('{}.*{}'.format(
    HYPERPARAMS_SECTION_START, HYPERPARAMS_SECTION_END), re.S)


class Hyperparameter(object):
    """Discrete grid hyperparameter.

    Represented by a list of values, a default value, the name of the
    hyperparameter (specified by the user in the workflow element), the
    name of the workflow element in which the hyperparemeter appears, and an
    optional prior probability vector.

    Attributes:
        name : string
            The name of the hyperparameter variable, used in user interface,
            both for specifying the grid of values and getting the report on
            an experiment. Initialized to '' then set in set_names, to the
            name the user chose for the variable in the workflow element.
        workflow_element_name : string
            The name of the workflow element in which the hyperparameter is
            used. Initialized to '' then set in set_names.
        dtype : string
            The dtype of the hyperparameter.
        default_index: int
            The index in values of the current value of the hyperparameter.
        values: numpy array of any dtype
            The list of hyperparameter values.
        prior: numpy array of float
            A list of reals that the hyperopt can use as a prior probability
            over values. Positivity and summing to one are not checked,
            hyperparameter optimizers should do that when using the list
    """

    def __init__(self, dtype, default=None, values=None, prior=None):
        self.name = ''
        self.workflow_element_name = ''
        self.dtype = dtype
        if default is None and values is None:
            raise ValueError('Either default or values must be defined.')
        if values is None:
            self.values = np.array([default], dtype=self.dtype)
        else:
            if len(values) < 1:
                raise ValueError(
                    'Values needs to contain at least one element.')
            self.values = np.array(values, dtype=self.dtype)
        if default is None:
            self.default_index = 0
        else:
            if default not in self.values:
                raise ValueError('Default must be among values.')
            else:
                self.set_default(default)

        if prior is None:
            self.prior = np.array([1. / self.n_values] * self.n_values)
        else:
            if len(prior) != len(values):
                raise ValueError(
                    'len(values) == {} != {} == len(prior)'.format(
                        len(values), len(prior)))
            self.prior = prior

    @property
    def n_values(self):
        """The number of hyperparameter values.

        Return:
            n_values : int
                The number of hyperparameter values len(values)
        """
        return len(self.values)

    @property
    def default(self):
        """The current value of the hyperparameter.

        Return:
            default : any dtype
                The current value of the hyperparameter values[default_index].
        """
        return self.values[self.default_index]

    @property
    def default_repr(self):
        """The string representation of the default value.

        It can be used to output the default value into a python file. For
        object types it adds '', otherwise it's the string representation of
        the default value.

        Return:
            default_repr : str
                The string representation of the default value.
        """
        if self.dtype == 'object':
            return '\'{}\''.format(self.default)
        else:
            return str(self.default)

    @property
    def values_repr(self):
        """The string representation of the list of values.

        It can be used to output the list of values into a python file. For
        object types it adds '' around the values, otherwise it's the list of
        string representations of the values in brackets.

        Return:
            values_repr : list of str
                The string representation of the list of values.
        """
        s = '['
        for v in self.values:
            if self.dtype == 'object':
                s += '\'{}\', '.format(v)
            else:
                s += '{}, '.format(v)
        s += ']'
        return s

    @property
    def python_repr(self):
        """The string representation of the hyperparameter.

        It can be used to output the hyperparameter definition into a python
        file:
        <name> = Hyperparameter(
            dtype=<dtype>, default=<default>, values=[<values>])

        Return:
            python_repr : str
                The string representation of the hyperparameter.
        """
        repr = '{} = Hyperparameter(\n'.format(self.name)
        repr += "\tdtype='{}'".format(str(self.dtype))
        repr += ', default={}'.format(self.default_repr)
        repr += ', values={})\n'.format(self.values_repr)
        return repr

    def set_names(self, name, workflow_element_name):
        """Set the name and workflow element name.

        Used when a hyperparameter object is loaded from a workflow element.

        Parameters:
            name : str
                The name of the hyperparameter, declared by the user in the
                workflow element.
            workflow_element_name : str
                The name of the workflow element in which the hyperparameter
                is defined.

        """
        self.name = name
        self.workflow_element_name = workflow_element_name

    def get_index(self, value):
        """Get the index of a value.

        Parameters:
            value : any dtype
                The value to look for.
        """
        return list(self.values).index(value)

    def set_default(self, default):
        """Set the default value.

        Parameters:
            default : any dtype
                The new default value.
        """
        self.default_index = self.get_index(default)

    def __int__(self):
        """Cast the default value into an integer.

        It can be used in the workflow element for an integer hyperparameter.

        Return:
            int(default) : int
                The integer representation of the default value.
        """
        return int(self.default)

    def __float__(self):
        """Cast the default value into an float.

        It can be used in the workflow element for an float hyperparameter.

        Return:
            float(default) : float
                The float representation of the default value.
        """
        return float(self.default)

    def __str__(self):
        """Cast the default value into a string.

        It can be used in the workflow element for a string hyperparameter.

        Return:
            str(default) : str
                The string representation of the default value.
        """
        return str(self.default)


def parse_hyperparameters(module_path, workflow_element_name):
    """Parse hyperparameters in a workflow element.

    Load the module, take all Hyperparameter objects, and set the name of each
    to the name of the hyperparameter the user chose and the workflow element
    name of each to workflow_element_name.

    Parameters:
        module_path : str
            The path to the submission directory.
        workflow_element_name : string
            The name of the workflow element.
    Return:
        hyperparameters : list of instances of Hyperparameter
    """
    hyperparameters = []
    workflow_element = import_module_from_source(
        os.path.join(module_path, workflow_element_name + '.py'),
        workflow_element_name
    )
    for object_name in dir(workflow_element):
        o = getattr(workflow_element, object_name)
        if type(o) == Hyperparameter:
            o.set_names(object_name, workflow_element_name)
            hyperparameters.append(o)
    return hyperparameters


def parse_all_hyperparameters(module_path, workflow):
    """Parse hyperparameters in a submission.

    Load all the the modules, take all Hyperparameter objects, and set the name
    of each to the name of the hyperparameter the user chose and the workflow
    element name of each to the corresponding workflow_element_name.

    Parameters:
        module_path : str
            The path to the submission directory.
        workflow_element_name : string
            The name of the workflow element.
    Return:
        hyperparameters : list of instances of Hyperparameter
    """
    hyperparameters = []
    for wen in workflow.element_names:
        hyperparameters += parse_hyperparameters(module_path, wen)
    return hyperparameters


def write_hyperparameters(submission_dir, output_submission_dir,
                          hypers_per_workflow_element):
    """Write hyperparameters in a submission.

    Read workflow elements from submission_dir, replace the hyperparameter
    section with the hyperparameters in the hypers_per_workflow_element
    dictionary (with new hyperparamter values set by, e.g, a hyperopt engine),
    then write the new workflow elements into output_submission_dir (which
    can be a temporary directory or submission_dir itself when the function
    is called to replace the hyperparameters in the input submission with the
    best hyperparameters.)

    Parameters:
        submission_dir : str
            The path to the submission directory from which the submission is
            read.
        output_submission_dir : str
            The path to the ouput submission directory into which the
            submission with the new hyperparameter values is written.
        hypers_per_workflow_element : dictionary
            Each key is a workflow element name and each value is a list of
            Hyperparameter instances, representing the hyperparemters in
            the workflow element.
    """
    for wen, hs in hypers_per_workflow_element.items():
        hyper_section = '{}\n'.format(HYPERPARAMS_SECTION_START)
        for h in hs:
            hyper_section += h.python_repr
        hyper_section += HYPERPARAMS_SECTION_END
        f_name = os.path.join(submission_dir, wen + '.py')
        with open(f_name) as f:
            content = f.read()
            content = HYPERPARAMS_REPL_REGEX.sub(hyper_section, content)
        output_f_name = os.path.join(output_submission_dir, wen + '.py')
        with open(output_f_name, 'w') as f:
            f.write(content)


class RandomEngine(object):
    """Random search hyperopt engine.

    Attributes:
        hyperparameters: a list of Hyperparameters
    """

    def __init__(self, hyperparameters):
        self.hyperparameters = hyperparameters

    def next_hyperparameter_indices(self, df_scores, n_folds):
        """Return the next hyperparameter indices to try.

        Parameters:
            df_scores : pandas DataFrame
                It represents the results of the experiments that have been
                run so far.
        Return:
            next_value_indices : list of int
                The indices in corresponding to the values to try in
                hyperparameters.
        """
        # First finish incomplete cv's.
        hyperparameter_names = [h.name for h in self.hyperparameters]
        df_n_folds = df_scores.groupby(hyperparameter_names).count()
        incomplete_folds = df_n_folds[(df_n_folds['fold_i'] % n_folds > 0)]
        if len(incomplete_folds) > 0:
            incomplete_folds = incomplete_folds.reset_index()
            next_values = incomplete_folds.iloc[0][
                [h.name for h in self.hyperparameters]].values
            next_value_indices = [
                h.get_index(v) for h, v
                in zip(self.hyperparameters, next_values)]
            # for some reason iloc converts int to float
            fold_i = int(incomplete_folds.iloc[0]['fold_i']) % n_folds
        # Otherwise select hyperparameter values from those that haven't
        # been selected yet, using also prior
        else:
            fold_i = 0
            next_value_indices = []
            df_scores_local = df_scores.copy()
            for h in self.hyperparameters:
                # unnormalized but positive prior
                prior = np.clip(h.prior, 1e-15, None)
                # How many times each hyperparameter value was tried, given the
                # selected values next_value_indices so far
                frequencies = np.zeros(len(prior))
                if len(df_scores_local) > 0:
                    for i, v in np.ndenumerate(h.values):
                        frequencies[i] = len(
                            df_scores_local[df_scores_local[h.name] == v])
                # How many times each hyperparameter value was not tried, given
                # the selected values next_value_indices so far, in this round
                # of full grid search
                frequencies = max(frequencies) - frequencies
                prior *= frequencies
                if prior.sum() <= 0:
                    prior = np.ones(len(prior))
                prior /= prior.sum()
                selected_index = np.random.choice(
                    range(len(h.values)), p=prior)
                # keep only experiments that used the selected values so far
                df_scores_local = df_scores_local[(
                    df_scores_local[h.name] == h.values[selected_index])]
                next_value_indices.append(selected_index)
        return fold_i, next_value_indices


class HyperparameterOptimization(object):
    """A hyperparameter optimization.

    Attributes:
        hyperparameters: a list of Hyperparameters
        engine: a hyperopt engine
        ramp_kit_dir: the directory where the ramp kit is found
        submission_dir: the directory where the submission to be optimized
            is found
    """

    def __init__(self, hyperparameters, engine, ramp_kit_dir, submission_dir):
        self.hyperparameters = hyperparameters
        self.engine = engine
        self.problem = assert_read_problem(ramp_kit_dir)
        self.X_train, self.y_train = self.problem.get_train_data(
            path=ramp_kit_dir)
        self.cv = list(self.problem.get_cv(self.X_train, self.y_train))
        self.submission_dir = submission_dir
        self.hyperparameter_names = [h.name for h in hyperparameters]
        self.score_names = [s.name for s in self.problem.score_types]
        self.df_summary_ = None

        # Set up hypers_per_workflow_element disctionary: keys are
        # workflow element names, values are lists are hypers belonging
        # to the workflow element
        self.hypers_per_workflow_element = {
            wen: [] for wen in self.problem.workflow.element_names}
        for h in self.hyperparameters:
            self.hypers_per_workflow_element[h.workflow_element_name].append(h)

        # Set up df_scores_ which will contain one row per experiment
        scores_columns = ['fold_i']
        scores_columns += self.hyperparameter_names
        scores_columns += ['train_' + name for name in self.score_names]
        scores_columns += ['valid_' + name for name in self.score_names]
        scores_columns += ['train_time', 'valid_time', 'n_train', 'n_valid']
        dtypes = ['int'] + [h.dtype for h in self.hyperparameters] +\
            ['float'] * 2 * len(self.score_names) + ['float'] * 2 + ['int'] * 2
        self.df_scores_ = pd.DataFrame(columns=scores_columns)
        for column, dtype in zip(scores_columns, dtypes):
            self.df_scores_[column] = self.df_scores_[column].astype(dtype)

    def _run_next_experiment(self, module_path, fold_i):
        _, _, df_scores = run_submission_on_cv_fold(
            self.problem, module_path=module_path, fold=self.cv[fold_i],
            X_train=self.X_train, y_train=self.y_train)
        return df_scores

    def _update_df_scores(self, df_scores, fold_i):
        row = {'fold_i': fold_i}
        for h in self.hyperparameters:
            row[h.name] = h.default
        for name in self.score_names:
            row['train_' + name] = df_scores.loc['train'][name]
            row['valid_' + name] = df_scores.loc['valid'][name]
        row['train_time'] = float(df_scores.loc['train']['time'])
        row['valid_time'] = float(df_scores.loc['valid']['time'])
        row['n_train'] = len(self.cv[fold_i][0])
        row['n_valid'] = len(self.cv[fold_i][1])
        self.df_scores_ = self.df_scores_.append(row, ignore_index=True)

    def _make_and_save_summary(self, hyperopt_output_path):
        summary_groupby = self.df_scores_.groupby(
            self.hyperparameter_names)
        means = summary_groupby.mean().drop(columns=['fold_i'])
        stds = summary_groupby.std().drop(columns=['fold_i'])
        counts = summary_groupby.count()[['n_train']].rename(
            columns={'n_train': 'n_folds'})
        self.df_summary_ = pd.merge(
            means, stds, left_index=True, right_index=True,
            suffixes=('_m', '_s'))
        self.df_summary_ = pd.merge(
            counts, self.df_summary_, left_index=True, right_index=True)
        print(self.df_summary_)
        summary_fname = os.path.join(hyperopt_output_path, 'summary.csv')
        self.df_summary_.to_csv(summary_fname)

    def _save_best_model(self):
        official_scores = self.df_summary_[
            'valid_' + self.problem.score_types[0].name + '_m']
        if self.problem.score_types[0].is_lower_the_better:
            best_defaults = official_scores.idxmin()
        else:
            best_defaults = official_scores.idxmax()
        print('Best hyperparameters: ', best_defaults)
        try:
            for bd, h in zip(best_defaults, self.hyperparameters):
                h.set_default(bd)
        except(TypeError):
            # single hyperparameter
            self.hyperparameters[0].set_default(best_defaults)
        # Overwrite the submission with the best hyperparameter values
        write_hyperparameters(
            self.submission_dir, self.submission_dir,
            self.hypers_per_workflow_element)

    def run(self, n_iter):
        # Create hyperopt output directory
        hyperopt_output_path = os.path.join(
            self.submission_dir, 'hyperopt_output')
        if not os.path.exists(hyperopt_output_path):
            os.makedirs(hyperopt_output_path)
        for i in range(n_iter):
            # Getting new hyperparameter values from engine
            fold_i, next_value_indices =\
                self.engine.next_hyperparameter_indices(
                    self.df_scores_, len(self.cv))
            # Updating hyperparameters
            for h, i in zip(self.hyperparameters, next_value_indices):
                h.default_index = i
            # Writing submission files with new hyperparameter values
            output_submission_dir = mkdtemp()
            write_hyperparameters(
                self.submission_dir, output_submission_dir,
                self.hypers_per_workflow_element)
            # Calling the training script.
            df_scores = self._run_next_experiment(
                output_submission_dir, fold_i)
            self._update_df_scores(df_scores, fold_i)
            shutil.rmtree(output_submission_dir)
        self._make_and_save_summary(hyperopt_output_path)
        self._save_best_model()


def init_hyperopt(ramp_kit_dir, ramp_submission_dir, submission, engine_name):
    problem = assert_read_problem(ramp_kit_dir)
    hyperopt_submission = submission + '_hyperopt'
    hyperopt_submission_dir = os.path.join(
        ramp_submission_dir, hyperopt_submission)
    submission_dir = os.path.join(
        ramp_submission_dir, submission)
    if os.path.exists(hyperopt_submission_dir):
        shutil.rmtree(hyperopt_submission_dir)
    shutil.copytree(submission_dir, hyperopt_submission_dir)
    hyperparameters = parse_all_hyperparameters(
        hyperopt_submission_dir, problem.workflow)
    if engine_name == 'random':
        engine = RandomEngine(hyperparameters)
    else:
        raise ValueError('{} is not a valide engine name'.format(engine_name))
    hyperparameter_experiment = HyperparameterOptimization(
        hyperparameters, engine, ramp_kit_dir, hyperopt_submission_dir)
    return hyperparameter_experiment


def run_hyperopt(ramp_kit_dir, ramp_data_dir, ramp_submission_dir,
                 submission, engine_name, n_iter, save_best=True,
                 is_cleanup=False):
    hyperparameter_experiment = init_hyperopt(
        ramp_kit_dir, ramp_submission_dir, submission, engine_name)
    hyperparameter_experiment.run(n_iter)
    if is_cleanup:
        shutil.rmtree(hyperparameter_experiment.submission_dir)



# coding: utf-8
"""
Utilities to manage the submissions
"""
import os
import time
import pickle
from collections.abc import Iterable
from collections import OrderedDict

import pandas as pd
import cloudpickle

from .io import save_y_pred, set_state, print_submission_exception
from .combine import get_score_cv_bags
from .pretty_print import print_title, print_df_scores, print_warning
from .scoring import score_matrix, round_df_scores, reorder_df_scores


def save_submissions(problem, y_pred, data_path='.', output_path='.',
                     suffix='test'):
    """Custom save a prediction vector in file to, e.g., submit to Kaggle.

    If problem.save_submission is implemented, y_pred is passed to it.
    Otherwise nothing happens (the exception is caught silently).

    Parameters
    ----------
    problem : a problem object
        loaded from problem.py, may implement save_y_pred
    y_pred : a prediction vector
        a vector of predictions to be saved
    data_path : str, (default='.')
        the directory of the ramp-kit to be tested for submission, maybe
        needed by problem.save_y_pred for, e.g., merging with an index vector
    output_path : str, (default='.')
        the directory where (typically) y_pred_<suffix>.csv will be saved
    suffix : str, (default='test')
        suffix in (typically) y_pred_<suffix>.csv, can be used in
        problem.save_y_pred to, e.g., save only test predictions
    """
    try:
        # We try using custom made problem.save_submission
        # it may need to re-read the data, e.g., for ids, so we send
        # it the data path. See, e.g.,
        # https://github.com/ramp-kits/kaggle_seguro/blob/master/problem.py
        problem.save_submission(y_pred, data_path, output_path, suffix)
    except AttributeError:
        pass


def train_test_submission(problem, module_path, X_train, y_train, X_test,
                          is_pickle, save_output, output_path,
                          model_name, train_is):
    """Train and test submission, on cv fold if train_is not none.

    Parameters
    ----------
    problem : problem object
        imp.loaded from problem.py
    module_path : str
        the path of the submission, typically submissions/<submission_name>
    X_train : a list of training instances
        returned by problem.get_train_data
    y_train : a list of training ground truth
        returned by problem.get_train_data
    X_test : a list of testing instances or None
        returned by problem.get_test_data
    is_pickle : boolean
        True if the model should be pickled
    output_path : str
        the path into which the model will be pickled
    save_output : boolean
        True if predictions should be written in files
    model_name : str (default='model.pkl')
        the file name of the pickled workflow
    train_is : a list of integers (default=None)
        training indices from the cross-validation fold, if None, train
        on full set
    Returns
    -------
    a tuple of the form ((y_pred_train, y_pred_test),
                         (train_time, valid_time, test_time))

    y_pred_train : a list of predictions
        on the training (train_train and train_valid) set
    y_pred_test : a list of predictions
        on the test set
    train_time : duration in seconds for training
    valid_time : duration in seconds for validation
    test_time : duration in seconds for testing
    """
    # Train
    t0 = time.time()
    try:
        trained_workflow = problem.workflow.train_submission(
            module_path, X_train, y_train, train_is)
    except Exception:
        print_submission_exception(save_output, output_path)
        set_state('training_error', save_output, output_path)
        exit(1)
    train_time = time.time() - t0
    set_state('trained', save_output, output_path)
    if is_pickle:
        trained_workflow = pickle_model(
            output_path, trained_workflow, model_name)

    # Validate
    t0 = time.time()
    try:
        y_pred_train = problem.workflow.test_submission(
            trained_workflow, X_train)
    except Exception:
        print_submission_exception(save_output, output_path)
        set_state('validating_error', save_output, output_path)
        exit(1)
    valid_time = time.time() - t0
    set_state('validated', save_output, output_path)

    # Test
    t0 = time.time()
    try:
        if X_test is None:
            y_pred_test = None
        else:
            y_pred_test = problem.workflow.test_submission(
                trained_workflow, X_test)
    except Exception:
        print_submission_exception(save_output, output_path)
        set_state('testing_error', save_output, output_path)
        exit(1)
    test_time = time.time() - t0
    set_state('tested', save_output, output_path)

    return (y_pred_train, y_pred_test), (train_time, valid_time, test_time)


def run_submission_on_cv_fold(problem, module_path, fold, X_train,
                              y_train, X_test=None, y_test=None,
                              is_pickle=False, save_output=False,
                              fold_output_path='.', ramp_data_dir='.'):
    """Run submission, compute and return predictions and scores on cv.

    Parameters
    ----------
    problem : problem object
        imp.loaded from problem.py
    module_path : str
        the path of the submission, typically submissions/<submission_name>
    X_train : a list of training instances
        returned by problem.get_train_data
    y_train : a list of training ground truth
        returned by problem.get_train_data
    X_train : a list of testing instances
        returned by problem.get_test_data
    y_test : a list of testing ground truth or None
        returned by problem.get_test_data
    is_pickle : boolean
        True if the model should be pickled
    save_output : boolean
        True if predictions should be written in files
    fold_output_path : str
        the path into which the model will be pickled
    fold : pair of lists of integers
        (train_is, valid_is) generated by problem.get_cv
    ramp_data_dir : str
        the directory of the data
    Returns
    -------
    predictions_train_valid : instance of Predictions
        on the validation set
    predictions_test : instance of Predictions
        on the test set
    df_scores : pd.DataFrame
        table of scores (rows = train/valid/test steps, columns = scores)
    """
    score_types = problem.score_types
    train_is, valid_is = fold
    pred, timing = train_test_submission(
        problem, module_path, X_train, y_train, X_test, is_pickle,
        save_output, fold_output_path, 'model.pkl', train_is)
    y_pred_train, y_pred_test = pred
    train_time, valid_time, test_time = timing

    predictions_train_train = problem.Predictions(
        y_pred=y_pred_train[train_is])
    ground_truth_train_train = problem.Predictions(
        y_true=y_train[train_is])
    predictions_train_valid = problem.Predictions(
        y_pred=y_pred_train[valid_is])
    ground_truth_train_valid = problem.Predictions(
        y_true=y_train[valid_is])
    if y_test is not None:
        predictions_test = problem.Predictions(y_pred=y_pred_test)
        ground_truth_test = problem.Predictions(y_true=y_test)
        if save_output:
            save_y_pred(
                problem, y_pred_train, data_path=ramp_data_dir,
                output_path=fold_output_path, suffix='train')
            if y_test is not None:
                save_y_pred(
                    problem, y_pred_test, data_path=ramp_data_dir,
                    output_path=fold_output_path, suffix='test')
            with open(os.path.join(fold_output_path, 'train_time'), 'w') as fd:
                fd.write(str(train_time))
            with open(os.path.join(fold_output_path, 'valid_time'), 'w') as fd:
                fd.write(str(valid_time))
            with open(os.path.join(fold_output_path, 'test_time'), 'w') as fd:
                fd.write(str(test_time))
        df_scores = score_matrix(
            score_types,
            ground_truth=OrderedDict([('train', ground_truth_train_train),
                                      ('valid', ground_truth_train_valid),
                                      ('test', ground_truth_test)]),
            predictions=OrderedDict([('train', predictions_train_train),
                                     ('valid', predictions_train_valid),
                                     ('test', predictions_test)]),
        )
        df_scores['time'] = [train_time, valid_time, test_time]
        set_state('scored', save_output, fold_output_path)
        return predictions_train_valid, predictions_test, df_scores

    else:
        if save_output:
            save_y_pred(
                problem, y_pred_train, data_path=ramp_data_dir,
                output_path=fold_output_path, suffix='train')
            with open(os.path.join(fold_output_path, 'train_time'), 'w') as fd:
                fd.write(str(train_time))
            with open(os.path.join(fold_output_path, 'valid_time'), 'w') as fd:
                fd.write(str(valid_time))
        df_scores = score_matrix(
            score_types,
            ground_truth=OrderedDict([('train', ground_truth_train_train),
                                      ('valid', ground_truth_train_valid)]),
            predictions=OrderedDict([('train', predictions_train_train),
                                     ('valid', predictions_train_valid)]),
        )
        df_scores['time'] = [train_time, valid_time]
        set_state('scored', save_output, fold_output_path)
        return predictions_train_valid, None, df_scores


def run_submission_on_full_train(problem, module_path, X_train, y_train,
                                 X_test, y_test, score_types,
                                 is_pickle, save_output, output_path,
                                 ramp_data_dir):
    """Run submission, compute predictions, and print scores on full train.

    Parameters
    ----------
    problem : problem object
        imp.loaded from problem.py
    module_path : str
        the path of the submission, typically submissions/<submission_name>
    X_train : a list of training instances
        returned by problem.get_train_data
    y_train : a list of training ground truth
        returned by problem.get_train_data
    X_test : a list of testing instances or None
        returned by problem.get_test_data
    y_test : a list of testing ground truth or None
        returned by problem.get_test_data
    score_types : a list of score types
        problem.score_types
    is_pickle : boolean
        True if the model should be pickled
    save_output : boolean
        True if predictions should be written in files
    output_path : str
        the path into which the model will be pickled
    ramp_data_dir : str
        the directory of the data
    """
    (y_pred_train, y_pred_test), _ = train_test_submission(
        problem, module_path, X_train, y_train, X_test, is_pickle,
        save_output, output_path, 'retrained_model.pkl', None)
    predictions_train = problem.Predictions(y_pred=y_pred_train)
    ground_truth_train = problem.Predictions(y_true=y_train)
    if y_test is not None:
        predictions_test = problem.Predictions(y_pred=y_pred_test)
        ground_truth_test = problem.Predictions(y_true=y_test)

        df_scores = score_matrix(
            score_types,
            ground_truth=OrderedDict([('train', ground_truth_train),
                                      ('test', ground_truth_test)]),
            predictions=OrderedDict([('train', predictions_train),
                                     ('test', predictions_test)]),
        )
        df_scores_rounded = round_df_scores(df_scores, score_types)
        print_df_scores(df_scores_rounded, indent='\t')

        if save_output:
            save_submissions(
                problem, y_pred_train, data_path=ramp_data_dir,
                output_path=output_path, suffix='retrain_train')
            save_submissions(
                problem, y_pred_test, data_path=ramp_data_dir,
                output_path=output_path, suffix='retrain_test')
    else:
        df_scores = score_matrix(
            score_types,
            ground_truth=OrderedDict([('train', ground_truth_train)]),
            predictions=OrderedDict([('train', predictions_train)]),
        )
        df_scores_rounded = round_df_scores(df_scores, score_types)
        print_df_scores(df_scores_rounded, indent='\t')

        if save_output:
            save_submissions(
                problem, y_pred_train, data_path=ramp_data_dir,
                output_path=output_path, suffix='retrain_train')


def bag_submissions(problem, cv, y_train, y_test, predictions_valid_list,
                    predictions_test_list, training_output_path,
                    ramp_data_dir='.', score_type_index=0,
                    save_output=False, score_table_title='Bagged scores',
                    score_f_name_prefix=''):
    """CV-bag trained submission.

    Parameters
    ----------
    problem : problem object
        imp.loaded from problem.py
    cv : cross validation object
        coming from get_cv of problem.py
    y_train : a list of training ground truth
        returned by problem.get_train_data
    y_test : a list of testing ground truth or None
        returned by problem.get_test_data
    predictions_valid_list : list of Prediction objects
        returned by run_submission_on_cv_fold
    predictions_test_list : list of Prediction objects or None
        returned by run_submission_on_cv_fold
    training_output_path : str
        submissions/<submission>/training_output
    ramp_data_dir : str
        the directory of the data
    score_type_index : int or None.
        The score type on which we bag. If None, all scores will be computed.
    save_output : boolean
        True if predictions should be written in files
    score_table_title : str
    score_f_name_prefix : str
    """
    print_title('----------------------------')
    print_title(score_table_title)
    print_title('----------------------------')
    score_type_index = (slice(None) if score_type_index is None
                        else score_type_index)
    score_types = problem.score_types[score_type_index]
    score_types = (
        [score_types] if not isinstance(score_types, Iterable)
        else score_types)

    # placeholder to store the scores and predictions
    bagged_scores = {}
    scoring_step = ['valid', 'test'] if y_test is not None else ['valid']
    for step in scoring_step:
        # Get either the training or testing infomation depending of the step
        pred_list = (predictions_valid_list if step == 'valid'
                     else predictions_test_list)
        y_step = y_train if step == 'valid' else y_test
        gt_list = problem.Predictions(y_true=y_step)
        # indices of the validation set or all sample for the testing set
        test_idx = ([valid_is for (train_is, valid_is) in cv]
                    if step == 'valid' else None)
        score_dict = {}
        for st in score_types:
            pred, scores = get_score_cv_bags(
                st, pred_list, gt_list, test_is_list=test_idx)
            score_dict[st.name] = {
                key: val for key, val in enumerate(scores)}
        bagged_scores[step] = score_dict
        # the predictions will always be the same for all score and we store
        # only a single instance
        if save_output:
            save_submissions(
                problem, pred.y_pred, data_path=ramp_data_dir,
                output_path=training_output_path,
                suffix='{}_bagged_{}'.format(score_f_name_prefix, step)
            )

    df_scores = pd.concat({step: pd.DataFrame(scores)
                           for step, scores in bagged_scores.items()})
    df_scores.columns = df_scores.columns.rename('score')
    df_scores.index = df_scores.index.rename(['step', 'n_bag'])
    # bagging learning curves can be plotted on this df_scores
    if save_output:
        bagged_scores_filename = os.path.join(
            training_output_path, 'bagged_scores.csv')
        df_scores.to_csv(bagged_scores_filename)

    # prepare the bagged scores which will be printed.
    highest_level = df_scores.index.get_level_values('n_bag').max()
    df_scores = df_scores.loc[(slice(None), highest_level), :]
    df_scores.index = df_scores.index.droplevel('n_bag')
    df_scores = reorder_df_scores(df_scores, score_types)
    df_scores = round_df_scores(df_scores, score_types)
    print_df_scores(df_scores, indent='\t')


def pickle_model(fold_output_path, trained_workflow, model_name='model.pkl'):
    """Pickle and reload trained workflow.

    If workflow can't be pickled, print warning and return original workflow.

    Parameters
    ----------
    fold_output_path : str
        the path into which the model will be pickled
    trained_workflow : a rampwf.workflow
        the workflow to be pickled
    model_name : str (default='model.pkl')
        the file name of the pickled workflow
    Returns
    -------
    trained_workflow : a rampwf.workflow
        either the input workflow or the pickled and reloaded workflow
    """
    msg = "Warning: model can't be pickled."
    model_file = os.path.join(fold_output_path, model_name)
    try:
        with open(model_file, 'wb') as pickle_file:
            cloudpickle.dump(trained_workflow, pickle_file)
    except pickle.PicklingError as e:
        print_warning(msg)
        print_warning(e)
        return trained_workflow
    else:
        # check if dumped trained_workflow can be loaded
        try:
            with open(model_file, 'rb') as pickle_file:
                trained_workflow = cloudpickle.load(pickle_file)
        except Exception as e:
            print_warning(msg)
            print_warning(e)

    return trained_workflow



from .feature_extractor import FeatureExtractor
from .classifier import Classifier


class FeatureExtractorClassifier(object):
    def __init__(self, workflow_element_names=[
            'feature_extractor', 'classifier']):
        self.element_names = workflow_element_names
        self.feature_extractor_workflow = FeatureExtractor(
            [self.element_names[0]])
        self.classifier_workflow = Classifier([self.element_names[1]])

    def train_submission(self, module_path, X_df, y_array, train_is=None):
        if train_is is None:
            train_is = slice(None, None, None)
        fe = self.feature_extractor_workflow.train_submission(
            module_path, X_df, y_array, train_is)
        X_train_array = self.feature_extractor_workflow.test_submission(
            fe, X_df.iloc[train_is])
        clf = self.classifier_workflow.train_submission(
            module_path, X_train_array, y_array[train_is])
        return fe, clf

    def test_submission(self, trained_model, X_df):
        fe, clf = trained_model
        X_test_array = self.feature_extractor_workflow.test_submission(
            fe, X_df)
        y_proba = self.classifier_workflow.test_submission(clf, X_test_array)
        return y_proba



"""
Train and test a grid feature extractor followed by a classifier.

This workflow is used to process gridded data where the input is
the spatial field at one time in X_ds matched with a corresponding label
in y_array. The dimensions of X_ds should be (time, y, x).
"""
from .grid_feature_extractor import GridFeatureExtractor
from .classifier import Classifier


class GridFeatureExtractorClassifier(object):
    def __init__(self, workflow_element_names=[
            'feature_extractor', 'classifier']):
        self.element_names = workflow_element_names
        self.feature_extractor_workflow = GridFeatureExtractor(
            [self.element_names[0]])
        self.classifier_workflow = Classifier([self.element_names[1]])

    def train_submission(self, module_path, X_ds, y_array, train_is=None):
        if train_is is None:
            train_is = slice(None, None, None)
        fe = self.feature_extractor_workflow.train_submission(
            module_path, X_ds, y_array, train_is)
        X_train_array = self.feature_extractor_workflow.test_submission(
            fe, X_ds.isel(enstime=train_is))
        clf = self.classifier_workflow.train_submission(
            module_path, X_train_array, y_array[train_is])
        return fe, clf

    def test_submission(self, trained_model, X_ds):
        fe, clf = trained_model
        X_test_array = self.feature_extractor_workflow.test_submission(
            fe, X_ds)
        y_proba = self.classifier_workflow.test_submission(clf, X_test_array)
        return y_proba



# coding: utf-8
"""
Utilities to deal with Jupyter notebooks
"""
from __future__ import print_function

import os
import sys


def delete_line_from_file(f_name, line_to_delete):
    with open(f_name, "r+", encoding='utf-8') as f:
        lines = f.readlines()
        f.seek(0)
        for line in lines:
            if line != line_to_delete:
                f.write(line)
        f.truncate()


def execute_notebook(ramp_kit_dir='.'):
    import nbformat
    from nbconvert.preprocessors import ExecutePreprocessor

    problem_name = os.path.basename(os.path.abspath(ramp_kit_dir))
    print('Testing if the notebook can be executed')
    notebook_filename = os.path.join(
        os.path.abspath(ramp_kit_dir),
        '{}_starting_kit.ipynb'.format(problem_name))
    kernel_name = 'python{}'.format(sys.version_info.major)

    with open(notebook_filename) as f:
        nb = nbformat.read(f, as_version=4)
        ep = ExecutePreprocessor(timeout=600, kernel_name=kernel_name)
        ep.preprocess(nb, {'metadata':
                           {'path': os.path.abspath(ramp_kit_dir)}})


def convert_notebook(ramp_kit_dir='.'):
    import nbformat
    import nbconvert
    from nbconvert.exporters import HTMLExporter

    problem_name = os.path.basename(os.path.abspath(ramp_kit_dir))
    print('Testing if the notebook can be converted to html')
    notebook_filename = os.path.join(
        os.path.abspath(ramp_kit_dir),
        '{}_starting_kit.ipynb'.format(problem_name))
    notebook_html_filename = os.path.join(
        os.path.abspath(ramp_kit_dir),
        '{}_starting_kit.html'.format(problem_name))

    with open(notebook_filename) as f:
        nb = nbformat.read(f, as_version=4)
        nb_html, _ = nbconvert.export(HTMLExporter, nb)

    with open(os.path.join(os.path.abspath(ramp_kit_dir),
                           notebook_html_filename), 'wb') as f:
        f.write(nb_html.encode('utf-8'))

    delete_line_from_file(
        notebook_html_filename,
        '<link rel="stylesheet" href="custom.css">\n')



from __future__ import division

import math


def cc_iou(circle1, circle2):
    """
    Intersection over Union (IoU) between two circles.

    Parameters
    ----------
    circle1 : tuple of floats
        first circle parameters (x_pos, y_pos, radius)
    circle2 : tuple of floats
        second circle parameters (x_pos, y_pos, radius)

    Returns
    -------
    float
        ratio between area of intersection and area of union

    """
    x1, y1, r1 = circle1
    x2, y2, r2 = circle2

    d = math.hypot(x2 - x1, y2 - y1)

    area_intersection = cc_intersection(d, r1, r2)
    area_union = math.pi * (r1 * r1 + r2 * r2) - area_intersection

    return area_intersection / area_union


def cc_intersection(dist, rad1, rad2):
    """
    Area of intersection between two circles.

    Parameters
    ----------
    dist : positive float
        distance between circle centers
    rad1 : positive float
        radius of first circle
    rad2 : positive float
        radius of second circle

    Returns
    -------
    intersection_area : positive float
        area of intersection between circles

    References
    ----------
    http://mathworld.wolfram.com/Circle-CircleIntersection.html

    """
    if dist < 0:
        raise ValueError("Distance between circles must be positive")
    if rad1 < 0 or rad2 < 0:
        raise ValueError("Circle radius must be positive")

    if dist == 0 or (dist <= abs(rad2 - rad1)):
        return min(rad1, rad2) ** 2 * math.pi

    if dist >= rad1 + rad2 or rad1 == 0 or rad2 == 0:
        return 0

    rad1_sq = rad1 * rad1
    rad2_sq = rad2 * rad2

    circle1 = rad1_sq * math.acos((dist * dist + rad1_sq - rad2_sq) /
                                  (2 * dist * rad1))
    circle2 = rad2_sq * math.acos((dist * dist + rad2_sq - rad1_sq) /
                                  (2 * dist * rad2))
    intersec = 0.5 * math.sqrt((-dist + rad1 + rad2) * (dist + rad1 - rad2) *
                               (dist - rad1 + rad2) * (dist + rad1 + rad2))
    intersection_area = circle1 + circle2 + intersec

    return intersection_area



import numpy as np
from .base import BaseScoreType


class RelativeRMSE(BaseScoreType):
    is_lower_the_better = True
    minimum = 0.0
    maximum = float('inf')

    def __init__(self, name='rmse', precision=2):
        self.name = name
        self.precision = precision

    def __call__(self, y_true, y_pred):
        return np.sqrt(np.mean(np.square((y_true - y_pred) / y_true)))



import os
import warnings

import click

from ..testing import assert_notebook
from ..testing import assert_submission

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


def get_submissions(ctx, args, incomplete):
    """Callback function for submission autocomplete.

    Find the possible submission names and use them for autocompletion.
    This only works if the submissions can be found in a 'submissions' folder.

    See click documentation for more information on the signature of the
    function.
    """
    submission_dir_path = os.path.join('.', 'submissions')
    try:
        all_submissions = [
            f.name for f in os.scandir(submission_dir_path)
            if f.is_dir() and f.name != '__pycache__']
    except FileNotFoundError:
        return []
    else:
        valid_submissions = [submission for submission in all_submissions
                             if incomplete in submission]
        return valid_submissions


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--submission', default='starting_kit', show_default=True,
              help='The kit to test. It should be located in the '
              '"submissions" folder of the starting kit. If "ALL", all '
              'submissions in the directory will be tested.',
              autocompletion=get_submissions)
@click.option('--ramp-kit-dir', default='.', show_default=True,
              help='Root directory of the ramp-kit to test.')
@click.option('--ramp-data-dir', default='.', show_default=True,
              help='Directory containing the data. This directory should '
              'contain a "data" folder.')
@click.option('--data-label', default=None, show_default=True,
              help='A label specifying the data in case the same submissions '
              'are executed on multiple datasets. If specified, '
              'problem.get_train_data and problem.get_test_data should '
              'accept a data_label argument. Typically they can deal with '
              'multiple datasets containing the data within the directory '
              'specified by --ramp-data-dir (default: ./data), for example '
              'using subdirectories ./data/<data_label>/. It is also '
              'the subdirectory of submissions/<submission>/training_output '
              'where results are saved if --save-output is used.')
@click.option('--ramp-submission-dir', default='submissions',
              show_default=True,
              help='Directory where the submissions are stored. It is the '
              'directory (typically called "submissions" in the ramp-kit) '
              'that contains the individual submission subdirectories.')
@click.option('--notebook', is_flag=True, show_default=True,
              help='Whether or not to test the notebook.')
@click.option('--quick-test', is_flag=True,
              help='Specify this flag to test the submission on a small '
              'subset of the data.')
@click.option('--pickle', is_flag=True,
              help='Specify this flag to pickle the submission after '
              'training.')
@click.option('--save-output', is_flag=True,
              help='Specify this flag to save predictions, scores, eventual '
              'error trace, and state after training.')
@click.option('--retrain', is_flag=True,
              help='Specify this flag to retrain the submission on the full '
              'training set after the CV loop.')
@click.option('--ignore-warning', is_flag=True,
              help='Will filters all warning and avoid to print them.')
def main(submission, ramp_kit_dir, ramp_data_dir, data_label,
         ramp_submission_dir, notebook, quick_test, pickle, save_output,
         retrain, ignore_warning):
    """Test a submission and/or a notebook before to submit on RAMP studio."""
    if quick_test:
        os.environ['RAMP_TEST_MODE'] = '1'

    if ignore_warning:
        warnings.simplefilter("ignore")

    if submission == "ALL":
        ramp_submission_dir = os.path.join(ramp_kit_dir, 'submissions')
        submission = [
            directory
            for directory in os.listdir(ramp_submission_dir)
            if os.path.isdir(os.path.join(ramp_submission_dir, directory))
        ]
    else:
        submission = [submission]

    for sub in submission:
        assert_submission(ramp_kit_dir=ramp_kit_dir,
                          ramp_data_dir=ramp_data_dir,
                          data_label=data_label,
                          ramp_submission_dir=ramp_submission_dir,
                          submission=sub,
                          is_pickle=pickle,
                          save_output=save_output,
                          retrain=retrain)

    if notebook:
        assert_notebook(ramp_kit_dir=ramp_kit_dir)


def start():
    main()


if __name__ == '__main__':
    start()



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
colored.py is a part of colored.

Copyright 2014-2017 Dimitris Zlatanidis <d.zlatanidis@gmail.com>
All rights reserved.

Colored is very simple Python library for color and formatting in terminal.

https://github.com/dslackw/colored

colored is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
               _                    _
      ___ ___ | | ___  _ __ ___  __| |
     / __/ _ \| |/ _ \| "__/ _ \/ _` |
    | (_| (_) | | (_) | | |  __/ (_| |
     \___\___/|_|\___/|_|  \___|\__,_|

    Very simple Python library for color and formatting in terminal.
    Collection of color codes and names for 256 color terminal setups.
    The following is a list of 256 colors for Xterm, containing an example
    of the displayed color, Xterm Name, Xterm Number.
"""


from .hex import HEX


class colored(object):

    def __init__(self, color):

        self.ESC = "\x1b["
        self.END = "m"
        self.color = color

        if str(color).startswith("#"):
            self.HEX = HEX(color.lower())
        else:
            self.HEX = ""

        self.paint = {
            "black": "0",
            "red": "1",
            "green": "2",
            "yellow": "3",
            "blue": "4",
            "magenta": "5",
            "cyan": "6",
            "light_gray": "7",
            "dark_gray": "8",
            "light_red": "9",
            "light_green": "10",
            "light_yellow": "11",
            "light_blue": "12",
            "light_magenta": "13",
            "light_cyan": "14",
            "white": "15",
            "grey_0": "16",
            "navy_blue": "17",
            "dark_blue": "18",
            "blue_3a": "19",
            "blue_3b": "20",
            "blue_1": "21",
            "dark_green": "22",
            "deep_sky_blue_4a": "23",
            "deep_sky_blue_4b": "24",
            "deep_sky_blue_4c": "25",
            "dodger_blue_3": "26",
            "dodger_blue_2": "27",
            "green_4": "28",
            "spring_green_4": "29",
            "turquoise_4": "30",
            "deep_sky_blue_3a": "31",
            "deep_sky_blue_3b": "32",
            "dodger_blue_1": "33",
            "green_3a": "34",
            "spring_green_3a": "35",
            "dark_cyan": "36",
            "light_sea_green": "37",
            "deep_sky_blue_2": "38",
            "deep_sky_blue_1": "39",
            "green_3b": "40",
            "spring_green_3b": "41",
            "spring_green_2a": "42",
            "cyan_3": "43",
            "dark_turquoise": "44",
            "turquoise_2": "45",
            "green_1": "46",
            "spring_green_2b": "47",
            "spring_green_1": "48",
            "medium_spring_green": "49",
            "cyan_2": "50",
            "cyan_1": "51",
            "dark_red_1": "52",
            "deep_pink_4a": "53",
            "purple_4a": "54",
            "purple_4b": "55",
            "purple_3": "56",
            "blue_violet": "57",
            "orange_4a": "58",
            "grey_37": "59",
            "medium_purple_4": "60",
            "slate_blue_3a": "61",
            "slate_blue_3b": "62",
            "royal_blue_1": "63",
            "chartreuse_4": "64",
            "dark_sea_green_4a": "65",
            "pale_turquoise_4": "66",
            "steel_blue": "67",
            "steel_blue_3": "68",
            "cornflower_blue": "69",
            "chartreuse_3a": "70",
            "dark_sea_green_4b": "71",
            "cadet_blue_2": "72",
            "cadet_blue_1": "73",
            "sky_blue_3": "74",
            "steel_blue_1a": "75",
            "chartreuse_3b": "76",
            "pale_green_3a": "77",
            "sea_green_3": "78",
            "aquamarine_3": "79",
            "medium_turquoise": "80",
            "steel_blue_1b": "81",
            "chartreuse_2a": "82",
            "sea_green_2": "83",
            "sea_green_1a": "84",
            "sea_green_1b": "85",
            "aquamarine_1a": "86",
            "dark_slate_gray_2": "87",
            "dark_red_2": "88",
            "deep_pink_4b": "89",
            "dark_magenta_1": "90",
            "dark_magenta_2": "91",
            "dark_violet_1a": "92",
            "purple_1a": "93",
            "orange_4b": "94",
            "light_pink_4": "95",
            "plum_4": "96",
            "medium_purple_3a": "97",
            "medium_purple_3b": "98",
            "slate_blue_1": "99",
            "yellow_4a": "100",
            "wheat_4": "101",
            "grey_53": "102",
            "light_slate_grey": "103",
            "medium_purple": "104",
            "light_slate_blue": "105",
            "yellow_4b": "106",
            "dark_olive_green_3a": "107",
            "dark_green_sea": "108",
            "light_sky_blue_3a": "109",
            "light_sky_blue_3b": "110",
            "sky_blue_2": "111",
            "chartreuse_2b": "112",
            "dark_olive_green_3b": "113",
            "pale_green_3b": "114",
            "dark_sea_green_3a": "115",
            "dark_slate_gray_3": "116",
            "sky_blue_1": "117",
            "chartreuse_1": "118",
            "light_green_2": "119",
            "light_green_3": "120",
            "pale_green_1a": "121",
            "aquamarine_1b": "122",
            "dark_slate_gray_1": "123",
            "red_3a": "124",
            "deep_pink_4c": "125",
            "medium_violet_red": "126",
            "magenta_3a": "127",
            "dark_violet_1b": "128",
            "purple_1b": "129",
            "dark_orange_3a": "130",
            "indian_red_1a": "131",
            "hot_pink_3a": "132",
            "medium_orchid_3": "133",
            "medium_orchid": "134",
            "medium_purple_2a": "135",
            "dark_goldenrod": "136",
            "light_salmon_3a": "137",
            "rosy_brown": "138",
            "grey_63": "139",
            "medium_purple_2b": "140",
            "medium_purple_1": "141",
            "gold_3a": "142",
            "dark_khaki": "143",
            "navajo_white_3": "144",
            "grey_69": "145",
            "light_steel_blue_3": "146",
            "light_steel_blue": "147",
            "yellow_3a": "148",
            "dark_olive_green_3": "149",
            "dark_sea_green_3b": "150",
            "dark_sea_green_2": "151",
            "light_cyan_3": "152",
            "light_sky_blue_1": "153",
            "green_yellow": "154",
            "dark_olive_green_2": "155",
            "pale_green_1b": "156",
            "dark_sea_green_5b": "157",
            "dark_sea_green_5a": "158",
            "pale_turquoise_1": "159",
            "red_3b": "160",
            "deep_pink_3a": "161",
            "deep_pink_3b": "162",
            "magenta_3b": "163",
            "magenta_3c": "164",
            "magenta_2a": "165",
            "dark_orange_3b": "166",
            "indian_red_1b": "167",
            "hot_pink_3b": "168",
            "hot_pink_2": "169",
            "orchid": "170",
            "medium_orchid_1a": "171",
            "orange_3": "172",
            "light_salmon_3b": "173",
            "light_pink_3": "174",
            "pink_3": "175",
            "plum_3": "176",
            "violet": "177",
            "gold_3b": "178",
            "light_goldenrod_3": "179",
            "tan": "180",
            "misty_rose_3": "181",
            "thistle_3": "182",
            "plum_2": "183",
            "yellow_3b": "184",
            "khaki_3": "185",
            "light_goldenrod_2a": "186",
            "light_yellow_3": "187",
            "grey_84": "188",
            "light_steel_blue_1": "189",
            "yellow_2": "190",
            "dark_olive_green_1a": "191",
            "dark_olive_green_1b": "192",
            "dark_sea_green_1": "193",
            "honeydew_2": "194",
            "light_cyan_1": "195",
            "red_1": "196",
            "deep_pink_2": "197",
            "deep_pink_1a": "198",
            "deep_pink_1b": "199",
            "magenta_2b": "200",
            "magenta_1": "201",
            "orange_red_1": "202",
            "indian_red_1c": "203",
            "indian_red_1d": "204",
            "hot_pink_1a": "205",
            "hot_pink_1b": "206",
            "medium_orchid_1b": "207",
            "dark_orange": "208",
            "salmon_1": "209",
            "light_coral": "210",
            "pale_violet_red_1": "211",
            "orchid_2": "212",
            "orchid_1": "213",
            "orange_1": "214",
            "sandy_brown": "215",
            "light_salmon_1": "216",
            "light_pink_1": "217",
            "pink_1": "218",
            "plum_1": "219",
            "gold_1": "220",
            "light_goldenrod_2b": "221",
            "light_goldenrod_2c": "222",
            "navajo_white_1": "223",
            "misty_rose1": "224",
            "thistle_1": "225",
            "yellow_1": "226",
            "light_goldenrod_1": "227",
            "khaki_1": "228",
            "wheat_1": "229",
            "cornsilk_1": "230",
            "grey_100": "231",
            "grey_3": "232",
            "grey_7": "233",
            "grey_11": "234",
            "grey_15": "235",
            "grey_19": "236",
            "grey_23": "237",
            "grey_27": "238",
            "grey_30": "239",
            "grey_35": "240",
            "grey_39": "241",
            "grey_42": "242",
            "grey_46": "243",
            "grey_50": "244",
            "grey_54": "245",
            "grey_58": "246",
            "grey_62": "247",
            "grey_66": "248",
            "grey_70": "249",
            "grey_74": "250",
            "grey_78": "251",
            "grey_82": "252",
            "grey_85": "253",
            "grey_89": "254",
            "grey_93": "255",
        }

    def attribute(self):
        """Set or reset attributes"""

        paint = {
            "bold": self.ESC + "1" + self.END,
            1: self.ESC + "1" + self.END,
            "dim": self.ESC + "2" + self.END,
            2: self.ESC + "2" + self.END,
            "underlined": self.ESC + "4" + self.END,
            4: self.ESC + "4" + self.END,
            "blink": self.ESC + "5" + self.END,
            5: self.ESC + "5" + self.END,
            "reverse": self.ESC + "7" + self.END,
            7: self.ESC + "7" + self.END,
            "hidden": self.ESC + "8" + self.END,
            8: self.ESC + "8" + self.END,
            "reset": self.ESC + "0" + self.END,
            0: self.ESC + "0" + self.END,
            "res_bold": self.ESC + "21" + self.END,
            21: self.ESC + "21" + self.END,
            "res_dim": self.ESC + "22" + self.END,
            22: self.ESC + "22" + self.END,
            "res_underlined": self.ESC + "24" + self.END,
            24: self.ESC + "24" + self.END,
            "res_blink": self.ESC + "25" + self.END,
            25: self.ESC + "25" + self.END,
            "res_reverse": self.ESC + "27" + self.END,
            27: self.ESC + "27" + self.END,
            "res_hidden": self.ESC + "28" + self.END,
            28: self.ESC + "28" + self.END,
        }
        return paint[self.color]

    def foreground(self):
        """Print 256 foreground colors"""
        code = self.ESC + "38;5;"
        if str(self.color).isdigit():
            self.reverse_dict()
            color = self.reserve_paint[str(self.color)]
            return code + self.paint[color] + self.END
        elif self.color.startswith("#"):
            return code + str(self.HEX) + self.END
        else:
            return code + self.paint[self.color] + self.END

    def background(self):
        """Print 256 background colors"""
        code = self.ESC + "48;5;"
        if str(self.color).isdigit():
            self.reverse_dict()
            color = self.reserve_paint[str(self.color)]
            return code + self.paint[color] + self.END
        elif self.color.startswith("#"):
            return code + str(self.HEX) + self.END
        else:
            return code + self.paint[self.color] + self.END

    def reverse_dict(self):
        """reverse dictionary"""
        self.reserve_paint = dict(zip(self.paint.values(), self.paint.keys()))


def attr(color):
    """alias for colored().attribute()"""
    return colored(color).attribute()


def fg(color):
    """alias for colored().foreground()"""
    return colored(color).foreground()


def bg(color):
    """alias for colored().background()"""
    return colored(color).background()


def stylize(text, styles, reset=True):
    """conveniently styles your text as and resets ANSI codes at its end."""
    terminator = attr("reset") if reset else ""
    return "{}{}{}".format("".join(styles), text, terminator)


def _c0wrap(styles):
    """wrap a set of ANSI styles in C0 control codes for readline safety."""
    C0_SOH = '\x01'   # mark the beginning of nonprinting characters
    C0_STX = '\x02'   # mark the end of nonprinting characters
    return "{}{}{}".format(C0_SOH, "".join(styles), C0_STX)


def stylize_interactive(text, styles, reset=True):
    """stylize() variant that adds C0 control codes (SOH/STX) for readline
    safety."""
    # problem: readline includes bare ANSI codes in width calculations.
    # solution: wrap nonprinting codes in SOH/STX when necessary.
    # see: https://github.com/dslackw/colored/issues/5
    terminator = _c0wrap(attr("reset")) if reset else ""
    return "{}{}{}".format(_c0wrap(styles), text, terminator)



"""
Utility for basic santizing of user provided Python scripts.
False negatives are expected.
"""
import sys

BLACKLIST = [
    # subprocess module
    'subprocess', 'Popen',
    # shutil module
    'shutil', 'rmtree', 'copytree',
    # os module
    'fdopen', 'fchmod', 'fchown', 'ftruncate', 'open', 'listdir', 'scandir',
    'removedirs'
]


def _sanitize_input(code):
    """Sanitize the user provided Python code

    This is not intended to be failproof by any mean, but merely provide early
    warning / detections of users trying to temper with the RAMP board system.
    """

    for key in BLACKLIST:
        if key in code:
            msg = "forbidden key word {} detected in submission.".format(key)
            if 'ramp_database' in sys.modules:
                msg += (
                    ' Tempering with the RAMP server is strictly forbidden! '
                    'Trying to cheat during a competition can cause '
                    'exclusion, legal consequences and/or a 0 grade in a '
                    'teaching course. The system administrator has been '
                    'informed. Please stop.'
                )
            raise RuntimeError(msg)
    return code



"""Base wrapper of scikit-learn-style numpy array predictions.

Functionalities:
- Hiding implementation details of prediction types (e.g., number of
  dimensions, semantics of columns).
- Providing a numpy array view through ``y_pred``. The ``y_pred`` view is
  used in ``rampwf.score``s as input, and in the default implementation
  of combining
- Handling cross-validation slices through ``set_valid_in_train``.
- Handling nan's in CV bagging by ``valid_indexes``.
- Combining ``Prediction``s (for CV bagging and ensembling). The default is
  to take the (nan)mean of the ``y_pred``s, but it can be overridden in
  derived classes.
Derived classes should all implement ``Prediction`` because we implement
polymorphism through importing ``Prediction`` from the particular file.
``Prediction``s can be asymmetric: ground truth (``y_true``) and predictions
are all stored in these classes. All ``rampwf.score_type``s accept
ground truth as the first argument and prediction as the second.
When constructors are called with a shape, they make an empty ``Prediction``.
This is to store combined ``Prediction``s, so the shape should be the same
as predictions' shape (in case of asymmetric ground truth/prediction).
"""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause

import numpy as np
import warnings


class BasePrediction(object):
    def __str__(self):
        return 'y_pred = {}'.format(self.y_pred)

    @property
    def valid_indexes(self):
        """Return valid indices (e.g., a cross-validation slice)."""
        if len(self.y_pred.shape) == 1:
            return ~np.isnan(self.y_pred)
        elif len(self.y_pred.shape) == 2:
            return ~np.isnan(self.y_pred[:, 0])
        else:
            raise ValueError('y_pred.shape > 2 is not implemented')

    def set_valid_in_train(self, predictions, test_is):
        """Set a cross-validation slice."""
        self.y_pred[test_is] = predictions.y_pred

    def check_y_pred_dimensions(self):
        if self.n_columns == 0 and len(self.y_pred.shape) != 1:
            raise ValueError(
                'Wrong y_pred dimensions: y_pred should be 1D, '
                'instead its shape is {}'.format(self.y_pred.shape))
        if self.n_columns > 0:
            if len(self.y_pred.shape) != 2 or\
                    self.y_pred.shape[1] != self.n_columns:
                raise ValueError(
                    'Wrong y_pred dimensions: y_pred should be 2D '
                    'with {} columns, instead its shape is {}'.format(
                        self.n_columns, self.y_pred.shape))

    @classmethod
    def combine(cls, predictions_list, index_list=None):
        """Combine predictions in predictions_list[index_list].

        The default implemented here is by taking the mean of their y_pred
        views. It can be overridden in derived classes.

        E.g. for regression it is the actual
        predictions, and for classification it is the probability array (which
        should be calibrated if we want the best performance). Called both for
        combining one submission on cv folds (a single model that is trained on
        different folds) and several models on a single fold.

        Parameters
        ----------
        predictions_list : list of instances of Base
            Each element of the list is an instance of Base with the
            same length and type.
        index_list : None | list of integers
            The subset of predictions to be combined. If None, the full set is
            combined.

        Returns
        -------
        combined_predictions : instance of cls
            A predictions instance containing the combined predictions.
        """
        if index_list is None:  # we combine the full list
            index_list = range(len(predictions_list))
        y_comb_list = np.array(
            [predictions_list[i].y_pred for i in index_list])
        # I expect to see RuntimeWarnings in this block
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=RuntimeWarning)
            y_comb = np.nanmean(y_comb_list, axis=0)
        combined_predictions = cls(y_pred=y_comb)
        return combined_predictions



"""Multiclass predictions.

``y_pred`` should be two dimensional (n_samples x n_classes).

"""
import itertools

import numpy as np
from scipy import sparse

from .base import BasePrediction
from ..score_types.detection.iou import cc_iou
from ..score_types.detection.util import _match_tuples


class Predictions(BasePrediction):

    iou_threshold = 0.5

    def __init__(self, y_pred=None, y_true=None, n_samples=None):
        if y_pred is not None:
            self.y_pred = y_pred
        elif y_true is not None:
            self.y_pred = y_true
        elif n_samples is not None:
            self.y_pred = np.empty(n_samples, dtype=object)
        else:
            raise ValueError(
                'Missing init argument: y_pred, y_true, or n_samples')
        self.check_y_pred_dimensions()

    def check_y_pred_dimensions(self):
        # XXX should check that prediction is an array of lists or Nones.
        pass

    @classmethod
    def combine(cls, predictions_list, index_list=None, greedy=False):

        if index_list is None:  # we combine the full list
            index_list = range(len(predictions_list))
        y_comb_list = [predictions_list[i].y_pred for i in index_list]

        n_preds = len(y_comb_list[0])

        y_preds_combined = np.empty(n_preds, dtype=object)

        for i in range(n_preds):

            preds_list = [preds[i] for preds in y_comb_list
                          if preds[i] is not None]

            if len(preds_list) == 1:
                # no overlap in the different prediction sets -> simply take
                # the single one that is not None
                preds_combined = preds_list[0]
            else:
                preds_combined, _ = combine_predictions(
                    preds_list, cls.iou_threshold, greedy=greedy)

            y_preds_combined[i] = preds_combined

        combined_predictions = cls(y_pred=y_preds_combined)
        return combined_predictions

    def set_valid_in_train(self, predictions, test_is):
        """Set a cross-validation slice."""
        self.y_pred[test_is] = predictions.y_pred

    @property
    def valid_indexes(self):
        """Return valid indices (e.g., a cross-validation slice)."""
        return self.y_pred != np.empty(len(self.y_pred), dtype=np.object)


def make_detection():
    return Predictions


def _greedy_nms(y_pred, iou_threshold=0.45):
    y_pred = np.asarray(y_pred)

    boxes_left = np.copy(y_pred)
    # This is where we store the boxes that make it through the
    # non-maximum suppression
    maxima = []

    # While there are still boxes left to compare...
    while boxes_left.shape[0] > 0:
        # ...get the index of the next box with the highest confidence...
        maximum_index = np.argmax(boxes_left[:, 0])
        # ...copy that box and...
        maximum_box = np.copy(boxes_left[maximum_index])
        # ...append it to `maxima` because we'll definitely keep it
        maxima.append(maximum_box)
        # Now remove the maximum box from `boxes_left`
        boxes_left = np.delete(boxes_left, maximum_index, axis=0)
        # If there are no boxes left after this step, break. Otherwise...
        if boxes_left.shape[0] == 0:
            break
        # ...compare (IoU) the other left over boxes to the maximum box...
        similarities = np.array([cc_iou(b[1:], maximum_box[1:]) for b in
                                 boxes_left])
        # ...so that we can remove the ones that overlap too much
        # with the maximum box
        boxes_left = boxes_left[similarities <= iou_threshold]

    return np.array(maxima)


def create_adjacency_matrix_from_edge_list(nodes, matches):
    # code based on nx.to_scipy_sparse_matrix

    nlen = len(nodes)
    index = dict(zip(nodes, range(nlen)))

    row, col, data = zip(*((index[u], index[v], 1)
                         for u, v in matches
                         if u in index and v in index))

    # symmetrize matrix
    d = data + data
    r = row + col
    c = col + row

    M = sparse.coo_matrix((d, (r, c)), shape=(nlen, nlen), dtype='int8')

    return M


def get_connected_components(nodes, matrix):
    from scipy.sparse.csgraph import connected_components
    ncon, labels = connected_components(matrix, directed=False)
    a_nodes = np.empty(len(nodes), dtype=object)
    a_nodes[:] = nodes
    return [a_nodes[labels == i] for i in range(ncon)]


def combine_predictions(preds_list, iou_threshold, greedy=False):
    """
    Combine multiple sets of predictions (of different models)
    for a single patch.


    """
    n_preds = len(preds_list)

    # for all pairs of the predictions: combine its detections and keep
    # those combinations that have a IOU > threshold
    matches = []
    matches_combined = []

    for mod1, mod2 in itertools.combinations(range(len(preds_list)), 2):

        pred1 = preds_list[mod1]
        pred2 = preds_list[mod2]

        idx1, idx2, ious = _match_tuples(
            [(x, y, r) for (c, x, y, r) in pred1],
            [(x, y, r) for (c, x, y, r) in pred2])

        idx1 = idx1[ious > iou_threshold]
        idx2 = idx2[ious > iou_threshold]

        # add each matching combination to list as (model_idx, pred_idx)
        # where:
        # model_idx is the index into preds_list
        # pred_idx is the index into preds_list[model_idx]
        for i1, i2 in zip(idx1, idx2):
            comb = (np.asarray(pred1[i1]) + np.array(pred2[i2])) / 2
            matches.append(((mod1, i1), (mod2, i2)))
            matches_combined.append(comb)

    if greedy:
        combined = _greedy_nms(matches_combined)
        return combined, matches

    # if matches is empty, return directly
    if not matches:
        return [], matches

    # the matches (set of (model_idx, pred_idx) tuples) are the nodes of
    # our graph:
    # - TODO: remove redundant edges/nodes
    # - convert to adjacency matrix
    # - get connected components of matrix
    nodes = sorted(set([x for y in matches for x in y]))
    M = create_adjacency_matrix_from_edge_list(nodes, matches)
    match_groups = get_connected_components(nodes, M)

    preds_combined = []

    # each connected set of detections is combined into a single (c,x,y,r)
    for group in match_groups:

        # convert list of (model_idx, pred_idx) tuples into actual
        # predictions
        preds = []
        for mod, idx in group:
            preds.append(preds_list[mod][idx])

        preds = np.array(preds)

        # combine in a single (c, x, y, r)
        if preds[:, 0].sum() == 0:
            # corner case of all confidence being zero: just taking mean
            # because weighted average will fail
            pred_combined = np.mean(preds, axis=0)
        else:
            # (x, y, r) are averaged weighted with the confidence
            pred_combined = np.average(preds[:, 1:], weights=preds[:, 0],
                                       axis=0)

            # the confidence is averaged taking into account missing
            # predictions
            conf = preds[:, 0].sum() / n_preds
            pred_combined = np.insert(pred_combined, 0, conf)

        preds_combined.append(pred_combined)

    return preds_combined, matches



import numpy as np
from .classifier_base import ClassifierBaseScoreType


def gini(y_true, y_pred, cmpcol=0, sortcol=1):
    all = np.asarray(
        np.c_[y_true, y_pred, np.arange(len(y_true))], dtype=np.float)
    all = all[np.lexsort((all[:, 2], -all[:, 1]))]
    total_losses = all[:, 0].sum()
    giniSum = all[:, 0].cumsum().sum() / total_losses
    giniSum -= (len(y_true) + 1) / 2.
    return giniSum / len(y_true)


class NormalizedGini(ClassifierBaseScoreType):
    is_lower_the_better = False
    minimum = -1.0
    maximum = 1.0

    def __init__(self, name='normalized_gini', precision=2):
        self.name = name
        self.precision = precision

    def score_function(self, ground_truths, predictions, valid_indexes=None):
        if valid_indexes is None:
            valid_indexes = slice(None, None, None)
        y_proba = predictions.y_pred[valid_indexes][:, 1]
        y_true_proba = ground_truths.y_pred_label_index[valid_indexes]
        self.check_y_pred_dimensions(y_true_proba, y_proba)
        return self.__call__(y_true_proba, y_proba)

    def __call__(self, y_true, y_pred):
        return gini(y_true, y_pred) / gini(y_true, y_true)



class BaseScoreType(object):
    def check_y_pred_dimensions(self, y_true, y_pred):
        if len(y_true) != len(y_pred):
            raise ValueError(
                'Wrong y_pred dimensions: y_pred should have {} instances, '
                'instead it has {} instances'.format(len(y_true), len(y_pred)))

    @property
    def worst(self):
        if self.is_lower_the_better:
            return self.maximum
        else:
            return self.minimum

    def score_function(self, ground_truths, predictions, valid_indexes=None):
        if valid_indexes is None:
            valid_indexes = slice(None, None, None)
        y_true = ground_truths.y_pred[valid_indexes]
        y_pred = predictions.y_pred[valid_indexes]
        self.check_y_pred_dimensions(y_true, y_pred)
        return self.__call__(y_true, y_pred)



"""Balanced accuracy.

From https://github.com/ch-imad/AutoMl_Challenge/blob/2353ec0/Starting_kit/scoring_program/libscores.py#L187  # noqa

See the thread at 
https://github.com/rhiever/tpot/issues/108#issuecomment-317067760
about the different definitions.
"""
from .classifier_base import ClassifierBaseScoreType
from .macro_averaged_recall import MacroAveragedRecall


class BalancedAccuracy(ClassifierBaseScoreType):
    is_lower_the_better = False
    minimum = 0.0
    maximum = 1.0

    def __init__(self, name='balanced_accuracy', precision=2):
        self.name = name
        self.precision = precision

    def __call__(self, y_true_label_index, y_pred_label_index):
        mac = MacroAveragedRecall()
        tpr = mac(y_true_label_index, y_pred_label_index)
        base_tpr = 1. / len(self.label_names)
        score = (tpr - base_tpr) / (1 - base_tpr)
        return score



# coding: utf-8

"""Provide utils to test ramp-kits."""
from __future__ import print_function

import os
import shutil

import numpy as np
import pandas as pd

from .combine import blend_on_fold
from .io import load_y_pred
from .importing import import_module_from_source
from .pretty_print import print_title, print_df_scores
from .notebook import execute_notebook, convert_notebook
from .scoring import round_df_scores, mean_score_matrix
from .submission import (bag_submissions, run_submission_on_cv_fold,
                         run_submission_on_full_train)


def assert_notebook(ramp_kit_dir='.'):
    print('----------------------------')
    convert_notebook(ramp_kit_dir)
    execute_notebook(ramp_kit_dir)


def assert_read_problem(ramp_kit_dir='.'):
    # giving a random name to the module so it passes looped tests
    return import_module_from_source(
        os.path.join(ramp_kit_dir, 'problem.py'), 'problem'
    )


def assert_title(ramp_kit_dir='.'):
    problem = assert_read_problem(ramp_kit_dir)
    print_title('Testing {}'.format(problem.problem_title))


def assert_data(ramp_kit_dir='.', ramp_data_dir='.', data_label=None):
    problem = assert_read_problem(ramp_kit_dir)
    print_title('Reading train and test files from {}/data/{} ...'.format(
        ramp_data_dir, f'{data_label}/' if data_label is not None else ""))
    kwargs = {}
    if data_label is not None:
        kwargs['data_label'] = data_label
    X_train, y_train = problem.get_train_data(path=ramp_data_dir, **kwargs)
    X_test, y_test = problem.get_test_data(path=ramp_data_dir, **kwargs)
    return X_train, y_train, X_test, y_test


def assert_cv(ramp_kit_dir='.', ramp_data_dir='.', data_label=None):
    problem = assert_read_problem(ramp_kit_dir)
    kwargs = {}
    if data_label is not None:
        kwargs['data_label'] = data_label
    X_train, y_train = problem.get_train_data(path=ramp_data_dir, **kwargs)
    print_title('Reading cv ...')
    cv = list(problem.get_cv(X_train, y_train))
    return cv


def assert_score_types(ramp_kit_dir='.'):
    problem = assert_read_problem(ramp_kit_dir)
    score_types = problem.score_types
    return score_types


def assert_submission(ramp_kit_dir='.', ramp_data_dir='.',
                      ramp_submission_dir='submissions', data_label=None,
                      submission='starting_kit', is_pickle=False,
                      save_output=False, retrain=False):
    """Helper to test a submission from a ramp-kit.

    Parameters
    ----------
    ramp_kit_dir : str, default='.'
        The directory of the ramp-kit to be tested for submission.
    ramp_data_dir : str, default='.'
        The directory of the data.
    ramp_submission_dir : str, default='./submissions'
        The directory of the submissions.
    data_label : str, default=None
        The subdirectory of data in /data and training outputs in
        /submissions/<submission>/training_output
    submission : str, default='starting_kit'
        The name of the submission to be tested.
    is_pickle : bool, default is False
        Whether to pickle the model or not.
    save_y_preds : bool, default is False
        Whether to store the predictions.
    retrain : bool, default is False
        Whether to train the model on the full training set and test on the
        test set.
    """
    problem = assert_read_problem(ramp_kit_dir)
    assert_title(ramp_kit_dir)
    X_train, y_train, X_test, y_test = assert_data(
        ramp_kit_dir, ramp_data_dir, data_label)
    cv = assert_cv(ramp_kit_dir, ramp_data_dir, data_label)
    score_types = assert_score_types(ramp_kit_dir)

    # module_path = os.path.join(ramp_kit_dir, 'submissions', submission)
    submission_path = os.path.join(ramp_submission_dir, submission)
    print_title('Training {} ...'.format(submission_path))

    training_output_path = ''
    if is_pickle or save_output:
        # creating <submission_path>/<submission>/training_output dir
        # optionally
        # <submission_path>/<submission>/training_output/<data_label>
        training_output_path = os.path.join(
            submission_path, 'training_output')
        if data_label is not None:
            training_output_path = os.path.join(
                training_output_path, data_label)
        if not os.path.exists(training_output_path):
            os.makedirs(training_output_path)

        print('Training output path: {}'.format(training_output_path))

    # saving predictions for CV bagging after the CV loop
    predictions_valid_list = []
    predictions_test_list = []
    df_scores_list = []

    for fold_i, fold in enumerate(cv):
        fold_output_path = ''
        if is_pickle or save_output:
            # creating <submission_path>/<submission>/training_output/fold_<i>
            fold_output_path = os.path.join(
                training_output_path, 'fold_{}'.format(fold_i))
            if not os.path.exists(fold_output_path):
                os.makedirs(fold_output_path)
        print_title('CV fold {}'.format(fold_i))

        predictions_valid, predictions_test, df_scores = \
            run_submission_on_cv_fold(
                problem, submission_path, fold, X_train, y_train,
                X_test, y_test, is_pickle, save_output, fold_output_path,
                ramp_data_dir)
        if save_output:
            filename = os.path.join(fold_output_path, 'scores.csv')
            df_scores.to_csv(filename)
        df_scores_rounded = round_df_scores(df_scores, score_types)
        print_df_scores(df_scores_rounded, indent='\t')

        # saving predictions for CV bagging after the CV loop
        df_scores_list.append(df_scores)
        predictions_valid_list.append(predictions_valid)
        predictions_test_list.append(predictions_test)

    print_title('----------------------------')
    print_title('Mean CV scores')
    print_title('----------------------------')
    df_mean_scores = mean_score_matrix(df_scores_list, score_types)
    print_df_scores(df_mean_scores, indent='\t')

    if retrain:
        # We retrain on the full training set
        print_title('----------------------------')
        print_title('Retrain scores')
        print_title('----------------------------')
        run_submission_on_full_train(
            problem, submission_path, X_train, y_train, X_test, y_test,
            score_types, is_pickle, save_output, training_output_path,
            ramp_data_dir)
    bag_submissions(
        problem, cv, y_train, y_test, predictions_valid_list,
        predictions_test_list, training_output_path,
        ramp_data_dir=ramp_data_dir, score_type_index=None,
        save_output=save_output)


def blend_submissions(submissions, ramp_kit_dir='.', ramp_data_dir='.',
                      ramp_submission_dir='.', data_label=None,
                      save_output=False, min_improvement=0.0):
    """Blending submissions in a ramp-kit and compute contributivities.

    If save_output is True, we create three files:
    <ramp_submission_dir>/training_output/contributivities.csv
    <ramp_submission_dir>/training_output/bagged_scores_combined.csv
    <ramp_submission_dir>/training_output/bagged_scores_foldwise_best.csv

    Parameters
    ----------
    submissions : list of str
        List of submission names (folders in <ramp_submission_dir>).
    ramp_kit_dir : str, default='.'
        The directory of the ramp-kit to be blended.
    ramp_data_dir : str, default='.'
        The directory of the data.
    data_label : str, default=None
        The subdirectory of data in {ramp_data_dir}/data and training outputs
        in {ramp_kit_dir}/submissions/<submission>/training_output
    ramp_submission_dir : str, default='./submissions'
        The directory of the submissions.
    save_output : bool, default is False
        Whether to store the blending results.
    min_improvement : float, default is 0.0
        The minimum improvement under which greedy blender is stopped.
    """
    problem = assert_read_problem(ramp_kit_dir)
    print_title('Blending {}'.format(problem.problem_title))
    X_train, y_train, X_test, y_test = assert_data(
        ramp_kit_dir, ramp_data_dir, data_label)
    cv = assert_cv(ramp_kit_dir, ramp_data_dir, data_label)
    valid_is_list = [valid_is for (train_is, valid_is) in cv]
    score_types = assert_score_types(ramp_kit_dir)
    n_folds = len(valid_is_list)
    contributivitys = np.zeros((len(submissions), n_folds))

    combined_predictions_valid_list = []
    foldwise_best_predictions_valid_list = []
    combined_predictions_test_list = []
    foldwise_best_predictions_test_list = []
    for fold_i, valid_is in enumerate(valid_is_list):
        print_title('CV fold {}'.format(fold_i))
        ground_truths_valid = problem.Predictions(y_true=y_train[valid_is])
        predictions_valid_list = []
        predictions_test_list = []
        for submission in submissions:
            module_path = os.path.join(ramp_submission_dir, submission)
            training_output_path = os.path.join(
                module_path, 'training_output')
            if data_label is not None:
                training_output_path = os.path.join(
                    training_output_path, data_label)
            fold_output_path = os.path.join(
                training_output_path, 'fold_{}'.format(fold_i))
            y_pred_train = load_y_pred(
                problem, data_path=ramp_data_dir,
                input_path=fold_output_path, suffix='train')
            y_pred_test = load_y_pred(
                problem, data_path=ramp_data_dir,
                input_path=fold_output_path, suffix='test')
            predictions_valid = problem.Predictions(
                y_pred=y_pred_train[valid_is])
            predictions_valid_list.append(predictions_valid)
            predictions_test = problem.Predictions(y_pred=y_pred_test)
            predictions_test_list.append(predictions_test)

        best_index_list = blend_on_fold(
            predictions_valid_list, ground_truths_valid, score_types[0],
            min_improvement=min_improvement)

        # we share a unit of 1. among the contributive submissions
        unit_contributivity = 1. / len(best_index_list)
        for i in best_index_list:
            contributivitys[i, fold_i] += unit_contributivity

        combined_predictions_valid_list.append(
            problem.Predictions.combine(
                predictions_valid_list, best_index_list))
        foldwise_best_predictions_valid_list.append(
            predictions_valid_list[best_index_list[0]])
        combined_predictions_test_list.append(
            problem.Predictions.combine(
                predictions_test_list, best_index_list))
        foldwise_best_predictions_test_list.append(
            predictions_test_list[best_index_list[0]])

    contributivitys /= n_folds
    contributivitys_df = pd.DataFrame()
    contributivitys_df['submission'] = np.array(submissions)
    contributivitys_df['contributivity'] = np.zeros(len(submissions))
    for fold_i in range(n_folds):
        c_i = contributivitys[:, fold_i]
        contributivitys_df['fold_{}'.format(fold_i)] = c_i
        contributivitys_df['contributivity'] += c_i
    percentage_factor = 100 / contributivitys_df['contributivity'].sum()
    contributivitys_df['contributivity'] *= percentage_factor
    rounded = contributivitys_df['contributivity'].round().astype(int)
    contributivitys_df['contributivity'] = rounded
    contributivitys_df = contributivitys_df.sort_values(
        'contributivity', ascending=False)
    print(contributivitys_df.to_string(index=False))

    if save_output:
        training_output_path = os.path.join(
            ramp_submission_dir, 'training_output')
        if not os.path.exists(training_output_path):
            os.mkdir(training_output_path)
        if data_label is not None:
            training_output_path = os.path.join(
                training_output_path, data_label)
            if not os.path.exists(training_output_path):
                os.makedirs(training_output_path)
        contributivitys_df.to_csv(os.path.join(
            training_output_path, 'contributivities.csv'), index=False)

    # bagging the foldwise ensembles
    bag_submissions(
        problem, cv, y_train, y_test, combined_predictions_valid_list,
        combined_predictions_test_list, training_output_path,
        ramp_data_dir=ramp_data_dir, score_type_index=0,
        save_output=save_output, score_table_title='Combined bagged scores',
        score_f_name_prefix='foldwise_best')
    if save_output:
        shutil.move(
            os.path.join(training_output_path, 'bagged_scores.csv'),
            os.path.join(training_output_path, 'bagged_scores_combined.csv'))

    # bagging the foldwise best submissions
    bag_submissions(
        problem, cv, y_train, y_test, foldwise_best_predictions_valid_list,
        foldwise_best_predictions_test_list, training_output_path,
        ramp_data_dir=ramp_data_dir, score_type_index=0,
        save_output=save_output,
        score_table_title='Foldwise best bagged scores',
        score_f_name_prefix='combined')
    if save_output:
        shutil.move(
            os.path.join(training_output_path, 'bagged_scores.csv'),
            os.path.join(
                training_output_path, 'bagged_scores_foldwise_best.csv'))



from __future__ import division

import math

import numpy as np

import pytest

from rampwf.score_types.detection.scp import scp_single
from rampwf.score_types.detection.ospa import ospa, ospa_single
from rampwf.score_types import DetectionAveragePrecision
from rampwf.score_types.detection.precision_recall import precision, recall
from rampwf.score_types.detection.precision_recall import mad_center
from rampwf.score_types.detection.precision_recall import mad_radius
from rampwf.score_types.detection.iou import cc_iou, cc_intersection
from rampwf.score_types.detection.scp import project_circle, circle_maps
from rampwf.score_types.detection.average_precision import (
    precision_recall_curve_greedy)


x = [(1, 1, 1)]
x2 = [(1, 1, 2)]
y = [(1, 3, 1)]
z = x + y
minipatch = [0, 2, 0, 2]


def test_scp_single():
    shape = (10, 10)
    # Perfect match
    assert scp_single(x, x, shape) == (0, 1, 1)
    # No match
    assert scp_single(x, y, shape) == (2, 1, 1)
    assert scp_single(x, x2, shape)[0] > 0
    # 1 match, 1 miss
    assert scp_single(x, z, shape) == (1, 1, 2)
    # 1 empty, 1 not
    assert scp_single(x, [], shape) == (1, 1, 0)
    assert scp_single([], x, shape) == (1, 0, 1)
    # 2 empty arrays
    assert scp_single([], [], shape) == (0, 0, 0)

    # Both inside minipatch
    assert scp_single(x, x, shape, minipatch=minipatch) == (0, 1, 1)
    # One mismatch
    assert scp_single(x, y, shape, minipatch=minipatch) == (1, 1, 0)
    # One too big
    assert scp_single(x, z, shape, minipatch=minipatch) == (0, 1, 1)
    assert scp_single(x, [], shape, minipatch=minipatch) == (1, 1, 0)
    assert scp_single([], x, shape, minipatch=minipatch) == (1, 0, 1)
    assert scp_single([], [], shape, minipatch=minipatch) == (0, 0, 0)

    # object outside of the image is discarded, although it actually is a
    # bad detection
    scp_single(x, [(20, 20, 1)], shape) == (1, 1, 0)


def test_ospa_single():
    # Perfect match
    assert ospa_single(x, x) == (2, 2, 2)
    # No match
    assert ospa_single(x, y) == (0, 2, 2)
    # assert ospa_single(x, x2) > 0
    # Miss or wrong radius
    assert ospa_single(x, z) == (2, 2, 3)
    # An empty array with a non empty one
    assert ospa_single(x, []) == (0, 0, 1)
    assert ospa_single([], x) == (0, 0, 1)
    assert ospa_single(z, []) == (0, 0, 2)
    # Two empty arrays should match
    assert ospa_single([], []) == (0, 0, 0)

    assert ospa_single(x, x, minipatch=minipatch) == (2, 2, 2)
    assert ospa_single(x, y, minipatch=minipatch) == (0, 1, 1)
    assert ospa_single(x, z, minipatch=minipatch) == (2, 2, 2)
    assert ospa_single(x, z, minipatch=minipatch) == (2, 2, 2)
    assert ospa_single(x, [], minipatch=minipatch) == (0, 0, 1)
    assert ospa_single([], x, minipatch=minipatch) == (0, 0, 1)
    assert ospa_single([], [], minipatch=minipatch) == (0, 0, 0)


def test_ospa():
    # Perfect match
    assert ospa([x], [x]) == 1
    assert ospa([x, x], [x, x]) == 1
    assert ospa([x, x], [x, y]) == 0.5
    assert ospa([x, x], [x, z]) == 4 / 5
    # No match
    assert ospa([x], [y]) == 0
    # assert ospa([x], [x]2) > 0
    assert ospa([x], [z]) == 2 / 3
    # Miss or wrong radius
    # An empty array with a non empty one
    assert ospa([x], [[]]) == 0
    assert ospa([[]], [x]) == 0
    assert ospa([z], [[]]) == 0
    # Two empty arrays should match
    assert ospa([[]], [[]]) == 0

    assert ospa([x], [x], minipatch=minipatch) == 2 / 2
    assert ospa([x, x], [x, x], minipatch=minipatch) == (2 + 2) / (2 + 2)
    assert ospa([x, x], [x, y], minipatch=minipatch) == (2 + 0) / (2 + 1)
    assert ospa([x, x], [x, z], minipatch=minipatch) == (2 + 2) / (2 + 2)
    assert ospa([x, x], [x, []], minipatch=minipatch) == (2 + 0) / (2 + 1)
    assert ospa([x, []], [x, x], minipatch=minipatch) == (2 + 0) / (2 + 1)
    assert ospa([x, []], [x, []], minipatch=minipatch) == (2 + 0) / (2 + 0)


def test_precision_recall():
    # perfect match
    y_true = [[(1, 1, 1), (3, 3, 1)]]
    y_pred = [[(1, 1, 1), (3, 3, 1)]]
    assert precision(y_true, y_pred) == 1
    assert recall(y_true, y_pred) == 1
    assert mad_radius(y_true, y_pred) == 0
    assert mad_center(y_true, y_pred) == 0

    # partly perfect match
    y_true = [[(1, 1, 1), (3, 3, 1), (7, 7, 1), (9, 9, 1)]]
    y_pred = [[(1, 1, 1), (5, 5, 1)]]
    assert precision(y_true, y_pred) == 0.5
    assert recall(y_true, y_pred) == 0.25
    assert mad_radius(y_true, y_pred) == 0
    assert mad_center(y_true, y_pred) == 0

    # imperfect match
    y_true = [[(1, 1, 1), (3, 3, 1), (7, 7, 1), (9, 9, 1)]]
    y_pred = [[(1, 1.2, 1.2), (3, 3, 1)]]
    assert precision(y_true, y_pred) == 1
    assert recall(y_true, y_pred) == 0.5
    assert mad_radius(y_true, y_pred) == pytest.approx(0.1)
    assert mad_center(y_true, y_pred) == pytest.approx(0.1)

    # no match
    y_true = [[(1, 1, 1)]]
    y_pred = [[(3, 3, 1)]]
    assert precision(y_true, y_pred) == 0
    assert recall(y_true, y_pred) == 0
    assert math.isnan(mad_radius(y_true, y_pred))
    assert math.isnan(mad_center(y_true, y_pred))


def test_average_precision():
    ap = DetectionAveragePrecision()

    # perfect match
    y_true = [[(1, 1, 1), (3, 3, 1)]]
    y_pred = [[(1, 1, 1, 1), (1, 3, 3, 1)]]
    assert ap(y_true, y_pred) == 1

    # imperfect match
    y_true = [[(1, 1, 1), (3, 3, 1), (7, 7, 1), (9, 9, 1)]]
    y_pred = [[(1, 1, 1, 1), (1, 5, 5, 1)]]
    assert ap(y_true, y_pred) == 0.125

    y_true = [[(1, 1, 1), (3, 3, 1), (7, 7, 1), (9, 9, 1)]]
    y_pred = [[(1, 1, 1.2, 1.2), (1, 3, 3, 1)]]
    assert ap(y_true, y_pred) == 0.5

    # no match
    y_true = [[], []]
    y_pred = [[], []]
    assert ap(y_true, y_pred) == 0
    y_true = [[(1, 1, 1)], []]
    y_pred = [[], []]
    assert ap(y_true, y_pred) == 0

    # only empty predictions
    y_true = [[(1, 1, 1)]]
    y_pred = [[(1, 3, 3, 1)]]
    assert ap(y_true, y_pred) == 0

    # bigger example
    y_true = [[(1, 1, 1), (3, 3, 1)], [(1, 1, 1), (3, 3, 1)]]
    y_pred = [[(0.9, 1, 1, 1), (0.7, 5, 5, 1), (0.5, 8, 8, 1)],
              [(0.8, 1, 1, 1), (0.6, 3, 3, 1), (0.4, 5, 5, 1)]]

    conf, ps, rs = precision_recall_curve_greedy(y_true, y_pred)
    assert conf.tolist() == [0.9, 0.8, 0.7, 0.6, 0.5, 0.4]
    assert ps.tolist() == [1, 1, 2 / 3, 3 / 4, 3 / 5, 3 / 6]  # noqa
    assert rs.tolist() == [1 / 4, 2 / 4, 2 / 4, 3 / 4, 3 / 4, 3 / 4]  # noqa
    assert ap(y_true, y_pred) == 11 / 16  # 0.5 * 1 + 0.25 * 3/4 + 0.25 * 0


# # test circles


circle = (1, 1, 1)
x = [circle]


def test_project_circle():
    shape = (10, 10)
    assert project_circle(circle, image=np.zeros(shape)).shape == shape
    assert project_circle(circle, shape=shape).shape == shape

    classic = project_circle(circle, shape=shape)
    assert classic.min() == 0
    assert classic.max() == 1

    negative = project_circle(circle, shape=shape, negative=True)
    assert negative.min() == -1
    assert negative.max() == 0

    normalized = project_circle(circle, shape=shape, normalize=True)
    assert normalized.min() == 0
    assert normalized.sum() == 1

    normalized_neg = project_circle(circle, shape=shape,
                                    normalize=True, negative=True)
    assert normalized_neg.max() == 0
    assert normalized_neg.sum() == -1

    with pytest.raises(ValueError):
        project_circle(circle)
        project_circle(circle, image=None, shape=None)


def test_circle_map():
    shape = (10, 10)
    map_true, map_pred = circle_maps([], [], shape)
    assert map_true.max() == 0
    assert map_pred.max() == 0
    assert map_true.sum() == 0
    assert map_pred.sum() == 0
    map_true, map_pred = circle_maps(x, [], shape)
    assert map_true.sum() == 1
    assert map_pred.sum() == 0
    map_true, map_pred = circle_maps([], x, shape)
    assert map_true.sum() == 0
    assert map_pred.sum() == 1
    map_true, map_pred = circle_maps(x, x, shape)
    assert map_true.sum() == 1
    assert map_pred.sum() == 1


# # test IOU


def test_cc_iou():
    circle1 = (0, 0, 1)
    circle2 = (0, 4, 1)
    circle3 = (1, 1, 2)
    circle1_2 = (0, 0, 2)
    assert cc_iou(circle1, circle1) - 1 < 1e-6
    assert cc_iou(circle1, circle2) < 1e-6
    assert cc_iou(circle2, circle1) < 1e-6
    assert cc_iou(circle1, circle3) - math.pi < 1e-6
    assert cc_iou(circle3, circle1) - math.pi < 1e-6
    assert cc_iou(circle1_2, circle1) == 0.25
    assert cc_iou(circle1, circle1_2) == 0.25


def test_cc_intersection():
    # Zero distance
    assert cc_intersection(0, 1, 2) - 4 * math.pi < 1e-6

    # Zero radius
    assert cc_intersection(1, 0, 1) == 0
    assert cc_intersection(1, 1, 0) == 0

    # High distance
    assert cc_intersection(4, 1, 2) == 0

    # Classic test
    assert cc_intersection(1, 1, 2) - math.pi < 1e-6

    with pytest.raises(ValueError):
        cc_intersection(-1, 1, 1)
        cc_intersection(1, -1, 1)
        cc_intersection(1, 1, -1)

    # floating point corner case
    assert cc_intersection(11.0, 6.0, 4.999999999999999) == 0


def test_cc_intersection_completely_overlapping():
    cc_intersection(2, 4, 1) - math.pi < 1e-6



from .base import BaseScoreType
import numpy as np
from sklearn.metrics import brier_score_loss


class BrierScore(BaseScoreType):
    is_lower_the_better = True
    minimum = 0.0
    maximum = 1.0

    def __init__(self, name='brier_score', precision=3):
        self.name = name
        self.precision = precision

    def score_function(self, ground_truths, predictions, valid_indexes=None):
        """A hybrid score.

        It tests the the predicted _probability_ of the second class
        against the true _label index_ (which is 0 if the first label is the
        ground truth, and 1 if it is not, in other words, it is the
        tru probabilty of the second class). Thus we have to override the
        `Base` function here
        """
        if valid_indexes is None:
            valid_indexes = slice(None, None, None)
        y_proba = predictions.y_pred[valid_indexes][:, 1]
        y_true_proba = ground_truths.y_pred_label_index[valid_indexes]
        self.check_y_pred_dimensions(y_true_proba, y_proba)
        return self.__call__(y_true_proba, y_proba)

    def __call__(self, y_true_proba, y_proba):
        return brier_score_loss(y_true_proba, y_proba)


class BrierSkillScore(BaseScoreType):
    is_lower_the_better = False
    minimum = -1.0
    maximum = 1.0

    def __init__(self, name='brier_score', precision=3):
        self.name = name
        self.precision = precision

    def score_function(self, ground_truths, predictions, valid_indexes=None):
        """A hybrid score.

        It tests the the predicted _probability_ of the second class
        against the true _label index_ (which is 0 if the first label is the
        ground truth, and 1 if it is not, in other words, it is the
        tru probabilty of the second class). Thus we have to override the
        `Base` function here
        """
        if valid_indexes is None:
            valid_indexes = slice(None, None, None)
        y_proba = predictions.y_pred[valid_indexes][:, 1]
        y_true_proba = ground_truths.y_pred_label_index[valid_indexes]
        self.check_y_pred_dimensions(y_true_proba, y_proba)
        return self.__call__(y_true_proba, y_proba)

    def __call__(self, y_true_proba, y_proba):
        climo = np.ones(y_true_proba.size) * y_true_proba.mean()
        bs = brier_score_loss(y_true_proba, y_proba)
        bs_c = brier_score_loss(y_true_proba, climo)
        return 1 - bs / bs_c


class BrierScoreReliability(BaseScoreType):
    is_lower_the_better = True
    minimum = 0.0
    maximum = 1.0

    def __init__(self, name='brier_score', precision=3,
                 bins=np.arange(0, 1.2, 0.1)):
        self.name = name
        self.precision = precision
        self.bins = bins
        self.bin_centers = (bins[1:] - bins[:-1]) * 0.05
        self.bin_centers[self.bin_centers > 1] = 1
        self.bin_centers[self.bin_centers < 0] = 0

    def score_function(self, ground_truths, predictions, valid_indexes=None):
        """A hybrid score.

        It tests the the predicted _probability_ of the second class
        against the true _label index_ (which is 0 if the first label is the
        ground truth, and 1 if it is not, in other words, it is the
        tru probabilty of the second class). Thus we have to override the
        `Base` function here
        """
        if valid_indexes is None:
            valid_indexes = slice(None, None, None)
        y_proba = predictions.y_pred[valid_indexes][:, 1]
        y_true_proba = ground_truths.y_pred_label_index[valid_indexes]
        self.check_y_pred_dimensions(y_true_proba, y_proba)
        return self.__call__(y_true_proba, y_proba)

    def __call__(self, y_true_proba, y_proba):
        pos_obs_freq = np.histogram(
            y_proba[y_true_proba == 1], bins=self.bins)[0]
        fore_freq = np.histogram(y_proba, bins=self.bins)[0]
        pos_obs_rel_freq = np.zeros(pos_obs_freq.size)
        for p in range(pos_obs_rel_freq.size):
            if fore_freq[p] > 0:
                pos_obs_rel_freq[p] = pos_obs_freq[p] / fore_freq[p]
            else:
                pos_obs_rel_freq[p] = np.nan
        score = np.nansum(
            fore_freq * (self.bin_centers - pos_obs_rel_freq) ** 2)
        score /= float(y_proba.size)

        return score


class BrierScoreResolution(BaseScoreType):
    is_lower_the_better = False
    minimum = 0.0
    maximum = 1.0

    def __init__(self, name='brier_score', precision=3,
                 bins=np.arange(0, 1.2, 0.1)):
        self.name = name
        self.precision = precision
        self.bins = bins
        self.bin_centers = (bins[1:] - bins[:-1]) * 0.05
        self.bin_centers[self.bin_centers > 1] = 1
        self.bin_centers[self.bin_centers < 0] = 0

    def score_function(self, ground_truths, predictions, valid_indexes=None):
        """A hybrid score.

        It tests the the predicted _probability_ of the second class
        against the true _label index_ (which is 0 if the first label is the
        ground truth, and 1 if it is not, in other words, it is the
        tru probabilty of the second class). Thus we have to override the
        `Base` function here
        """
        if valid_indexes is None:
            valid_indexes = slice(None, None, None)
        y_proba = predictions.y_pred[valid_indexes][:, 1]
        y_true_proba = ground_truths.y_pred_label_index[valid_indexes]
        self.check_y_pred_dimensions(y_true_proba, y_proba)
        return self.__call__(y_true_proba, y_proba)

    def __call__(self, y_true_proba, y_proba):
        """
        See Murphy (1973) A vector partition of the probability score
        """
        np.seterr(divide="ignore")
        pos_obs_freq = np.histogram(
            y_proba[y_true_proba == 1], bins=self.bins)[0]
        fore_freq = np.histogram(y_proba, bins=self.bins)[0]
        climo = y_true_proba.mean()
        unc = climo * (1 - climo)
        pos_obs_rel_freq = np.zeros(pos_obs_freq.size)
        for p in range(pos_obs_rel_freq.size):
            if fore_freq[p] > 0:
                pos_obs_rel_freq[p] = pos_obs_freq[p] / fore_freq[p]
            else:
                pos_obs_rel_freq[p] = np.nan
        score = np.nansum(fore_freq * (pos_obs_rel_freq - climo) ** 2)
        score /= float(y_proba.size)
        return score / unc



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
colors.py is a part of colored.

Copyright 2014-2017 Dimitris Zlatanidis <d.zlatanidis@gmail.com>
All rights reserved.

Colored is very simple Python library for color and formatting in terminal.

https://github.com/dslackw/colored

colored is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""


names = [
    'BLACK',
    'RED',
    'GREEN',
    'YELLOW',
    'BLUE',
    'MAGENTA',
    'CYAN',
    'LIGHT_GRAY',
    'DARK_GRAY',
    'LIGHT_RED',
    'LIGHT_GREEN',
    'LIGHT_YELLOW',
    'LIGHT_BLUE',
    'LIGHT_MAGENTA',
    'LIGHT_CYAN',
    'WHITE',
    'GREY_0',
    'NAVY_BLUE',
    'DARK_BLUE',
    'BLUE_3A',
    'BLUE_3B',
    'BLUE_1',
    'DARK_GREEN',
    'DEEP_SKY_BLUE_4A',
    'DEEP_SKY_BLUE_4B',
    'DEEP_SKY_BLUE_4C',
    'DODGER_BLUE_3',
    'DODGER_BLUE_2',
    'GREEN_4',
    'SPRING_GREEN_4',
    'TURQUOISE_4',
    'DEEP_SKY_BLUE_3A',
    'DEEP_SKY_BLUE_3B',
    'DODGER_BLUE_1',
    'GREEN_3A',
    'SPRING_GREEN_3A',
    'DARK_CYAN',
    'LIGHT_SEA_GREEN',
    'DEEP_SKY_BLUE_2',
    'DEEP_SKY_BLUE_1',
    'GREEN_3B',
    'SPRING_GREEN_3B',
    'SPRING_GREEN_2A',
    'CYAN_3',
    'DARK_TURQUOISE',
    'TURQUOISE_2',
    'GREEN_1',
    'SPRING_GREEN_2B',
    'SPRING_GREEN_1',
    'MEDIUM_SPRING_GREEN',
    'CYAN_2',
    'CYAN_1',
    'DARK_RED_1',
    'DEEP_PINK_4A',
    'PURPLE_4A',
    'PURPLE_4B',
    'PURPLE_3',
    'BLUE_VIOLET',
    'ORANGE_4A',
    'GREY_37',
    'MEDIUM_PURPLE_4',
    'SLATE_BLUE_3A',
    'SLATE_BLUE_3B',
    'ROYAL_BLUE_1',
    'CHARTREUSE_4',
    'DARK_SEA_GREEN_4A',
    'PALE_TURQUOISE_4',
    'STEEL_BLUE',
    'STEEL_BLUE_3',
    'CORNFLOWER_BLUE',
    'CHARTREUSE_3A',
    'DARK_SEA_GREEN_4B',
    'CADET_BLUE_2',
    'CADET_BLUE_1',
    'SKY_BLUE_3',
    'STEEL_BLUE_1A',
    'CHARTREUSE_3B',
    'PALE_GREEN_3A',
    'SEA_GREEN_3',
    'AQUAMARINE_3',
    'MEDIUM_TURQUOISE',
    'STEEL_BLUE_1B',
    'CHARTREUSE_2A',
    'SEA_GREEN_2',
    'SEA_GREEN_1A',
    'SEA_GREEN_1B',
    'AQUAMARINE_1A',
    'DARK_SLATE_GRAY_2',
    'DARK_RED_2',
    'DEEP_PINK_4B',
    'DARK_MAGENTA_1',
    'DARK_MAGENTA_2',
    'DARK_VIOLET_1A',
    'PURPLE_1A',
    'ORANGE_4B',
    'LIGHT_PINK_4',
    'PLUM_4',
    'MEDIUM_PURPLE_3A',
    'MEDIUM_PURPLE_3B',
    'SLATE_BLUE_1',
    'YELLOW_4A',
    'WHEAT_4',
    'GREY_53',
    'LIGHT_SLATE_GREY',
    'MEDIUM_PURPLE',
    'LIGHT_SLATE_BLUE',
    'YELLOW_4B',
    'DARK_OLIVE_GREEN_3A',
    'DARK_GREEN_SEA',
    'LIGHT_SKY_BLUE_3A',
    'LIGHT_SKY_BLUE_3B',
    'SKY_BLUE_2',
    'CHARTREUSE_2B',
    'DARK_OLIVE_GREEN_3B',
    'PALE_GREEN_3B',
    'DARK_SEA_GREEN_3A',
    'DARK_SLATE_GRAY_3',
    'SKY_BLUE_1',
    'CHARTREUSE_1',
    'LIGHT_GREEN_2',
    'LIGHT_GREEN_3',
    'PALE_GREEN_1A',
    'AQUAMARINE_1B',
    'DARK_SLATE_GRAY_1',
    'RED_3A',
    'DEEP_PINK_4C',
    'MEDIUM_VIOLET_RED',
    'MAGENTA_3A',
    'DARK_VIOLET_1B',
    'PURPLE_1B',
    'DARK_ORANGE_3A',
    'INDIAN_RED_1A',
    'HOT_PINK_3A',
    'MEDIUM_ORCHID_3',
    'MEDIUM_ORCHID',
    'MEDIUM_PURPLE_2A',
    'DARK_GOLDENROD',
    'LIGHT_SALMON_3A',
    'ROSY_BROWN',
    'GREY_63',
    'MEDIUM_PURPLE_2B',
    'MEDIUM_PURPLE_1',
    'GOLD_3A',
    'DARK_KHAKI',
    'NAVAJO_WHITE_3',
    'GREY_69',
    'LIGHT_STEEL_BLUE_3',
    'LIGHT_STEEL_BLUE',
    'YELLOW_3A',
    'DARK_OLIVE_GREEN_3',
    'DARK_SEA_GREEN_3B',
    'DARK_SEA_GREEN_2',
    'LIGHT_CYAN_3',
    'LIGHT_SKY_BLUE_1',
    'GREEN_YELLOW',
    'DARK_OLIVE_GREEN_2',
    'PALE_GREEN_1B',
    'DARK_SEA_GREEN_5B',
    'DARK_SEA_GREEN_5A',
    'PALE_TURQUOISE_1',
    'RED_3B',
    'DEEP_PINK_3A',
    'DEEP_PINK_3B',
    'MAGENTA_3B',
    'MAGENTA_3C',
    'MAGENTA_2A',
    'DARK_ORANGE_3B',
    'INDIAN_RED_1B',
    'HOT_PINK_3B',
    'HOT_PINK_2',
    'ORCHID',
    'MEDIUM_ORCHID_1A',
    'ORANGE_3',
    'LIGHT_SALMON_3B',
    'LIGHT_PINK_3',
    'PINK_3',
    'PLUM_3',
    'VIOLET',
    'GOLD_3B',
    'LIGHT_GOLDENROD_3',
    'TAN',
    'MISTY_ROSE_3',
    'THISTLE_3',
    'PLUM_2',
    'YELLOW_3B',
    'KHAKI_3',
    'LIGHT_GOLDENROD_2A',
    'LIGHT_YELLOW_3',
    'GREY_84',
    'LIGHT_STEEL_BLUE_1',
    'YELLOW_2',
    'DARK_OLIVE_GREEN_1A',
    'DARK_OLIVE_GREEN_1B',
    'DARK_SEA_GREEN_1',
    'HONEYDEW_2',
    'LIGHT_CYAN_1',
    'RED_1',
    'DEEP_PINK_2',
    'DEEP_PINK_1A',
    'DEEP_PINK_1B',
    'MAGENTA_2B',
    'MAGENTA_1',
    'ORANGE_RED_1',
    'INDIAN_RED_1C',
    'INDIAN_RED_1D',
    'HOT_PINK_1A',
    'HOT_PINK_1B',
    'MEDIUM_ORCHID_1B',
    'DARK_ORANGE',
    'SALMON_1',
    'LIGHT_CORAL',
    'PALE_VIOLET_RED_1',
    'ORCHID_2',
    'ORCHID_1',
    'ORANGE_1',
    'SANDY_BROWN',
    'LIGHT_SALMON_1',
    'LIGHT_PINK_1',
    'PINK_1',
    'PLUM_1',
    'GOLD_1',
    'LIGHT_GOLDENROD_2B',
    'LIGHT_GOLDENROD_2C',
    'NAVAJO_WHITE_1',
    'MISTY_ROSE1',
    'THISTLE_1',
    'YELLOW_1',
    'LIGHT_GOLDENROD_1',
    'KHAKI_1',
    'WHEAT_1',
    'CORNSILK_1',
    'GREY_100',
    'GREY_3',
    'GREY_7',
    'GREY_11',
    'GREY_15',
    'GREY_19',
    'GREY_23',
    'GREY_27',
    'GREY_30',
    'GREY_35',
    'GREY_39',
    'GREY_42',
    'GREY_46',
    'GREY_50',
    'GREY_54',
    'GREY_58',
    'GREY_62',
    'GREY_66',
    'GREY_70',
    'GREY_74',
    'GREY_78',
    'GREY_82',
    'GREY_85',
    'GREY_89',
    'GREY_93'
    ]



import os

from ..utils.importing import import_module_from_source


class Classifier(object):
    def __init__(self, workflow_element_names=['classifier']):
        self.element_names = workflow_element_names
        # self.name = 'classifier_workflow'  # temporary

    def train_submission(self, module_path, X_array, y_array, train_is=None):
        if train_is is None:
            train_is = slice(None, None, None)
        classifier = import_module_from_source(
            os.path.join(module_path, self.element_names[0] + '.py'),
            self.element_names[0],
            sanitize=True
        )
        clf = classifier.Classifier()
        clf.fit(X_array[train_is], y_array[train_is])
        return clf

    def test_submission(self, trained_model, X_array):
        clf = trained_model
        y_proba = clf.predict_proba(X_array)
        return y_proba



import os
import pytest
from rampwf.hyperopt import Hyperparameter, run_hyperopt

PATH = os.path.dirname(__file__)


def test_hyperparameter():
    hp_1 = Hyperparameter(dtype='int', values=[1, 2, 3])
    assert hp_1.default == 1
    assert hp_1.n_values == 3
    assert list(hp_1.values) == [1, 2, 3]
    assert list(hp_1.prior) == [1. / 3, 1. / 3, 1. / 3]
    hp_2 = Hyperparameter(dtype='int', values=[1, 2, 3], prior=[0.1, 0.7, 0.2])
    assert hp_2.default == 1
    assert hp_2.n_values == 3
    assert list(hp_2.values) == [1, 2, 3]
    assert list(hp_2.prior) == [0.1, 0.7, 0.2]
    hp_3 = Hyperparameter(dtype='int', default=1)
    assert hp_3.default == 1
    assert hp_3.n_values == 1
    assert list(hp_3.values) == [1]
    assert list(hp_3.prior) == [1.0]
    with pytest.raises(ValueError) as e:
        Hyperparameter(dtype='int', values=[1, 2, 3], prior=[0.1, 0.7])
    assert str(e.value) == 'len(values) == 3 != 2 == len(prior)'
    with pytest.raises(ValueError) as e:
        Hyperparameter(dtype='int')
    assert str(e.value) == 'Either default or values must be defined.'
    with pytest.raises(ValueError) as e:
        Hyperparameter(dtype='int', values=[])
    assert str(e.value) == 'Values needs to contain at least one element.'
    with pytest.raises(ValueError) as e:
        Hyperparameter(dtype='int', default=2, values=[1])
    assert str(e.value) == 'Default must be among values.'


@pytest.mark.parametrize("submission", ['starting_kit', 'one_hyper_kit'])
def test_hyperopt(submission):
    ramp_kit_dir = os.path.join(
        PATH, 'interfaces', 'header_in_files', 'titanic')
    run_hyperopt(
        ramp_kit_dir, ramp_kit_dir,
        os.path.join(ramp_kit_dir, 'submissions'),
        submission, 'random', 64, is_cleanup=True)



"""Testing for mixed predictions (rampwf.prediction.mixed)."""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause

# import numpy as np
# from rampwf.prediction_types.mixed import Predictions
# from numpy.testing import assert_array_equal


# def test_init():
#     labels = ['Class_1', 'Class_2', 'Class_3']
#     ps_1 = [0.7, 0.1, 0.2, 1.85]
#     ps_2 = [0.1, 0.1, 0.8, 0.15]
#     ps_3 = [0.2, 0.5, 0.3, -0.22]
#     predictions = Predictions(
#         labels=labels, y_pred=np.array([ps_1, ps_2, ps_3]))
#     assert_array_equal(predictions.y_pred, [ps_1, ps_2, ps_3])
#     assert_array_equal(predictions.multiclass.y_pred_label_index, [0, 2, 1])
#     assert_array_equal(
#         predictions.multiclass.y_pred, [ps_1[:-1], ps_2[:-1], ps_3[:-1]])
#     assert_array_equal(
#         predictions.regression.y_pred, [ps_1[-1], ps_2[-1], ps_3[-1]])


# def test_init_empty():
#     labels = ['Class_1', 'Class_2', 'Class_3']
#     predictions = Predictions(labels=labels, shape=(4, 4))
#     assert_array_equal(predictions.y_pred, np.array([
#         [np.nan, np.nan, np.nan, np.nan],
#         [np.nan, np.nan, np.nan, np.nan],
#         [np.nan, np.nan, np.nan, np.nan],
#         [np.nan, np.nan, np.nan, np.nan],
#     ]))


# def test_set_valid_in_train():
#     labels = ['Class_1', 'Class_2', 'Class_3']
#     predictions = Predictions(labels=labels, shape=(4, 4))
#     assert_array_equal(predictions.y_pred, np.array([
#         [np.nan, np.nan, np.nan, np.nan],
#         [np.nan, np.nan, np.nan, np.nan],
#         [np.nan, np.nan, np.nan, np.nan],
#         [np.nan, np.nan, np.nan, np.nan],
#     ]))
#     ps_1 = [0.7, 0.1, 0.2, 1.85]
#     ps_2 = [0.1, 0.1, 0.8, 0.15]
#     ps_3 = [0.2, 0.5, 0.3, -0.22]
#     predictions_valid = Predictions(
#         labels=labels, y_pred=np.array([ps_1, ps_2, ps_3]))
#     predictions.set_valid_in_train(predictions_valid, [0, 2, 3])
#     assert_array_equal(predictions.y_pred, np.array([
#         [0.7, 0.1, 0.2, 1.85],
#         [np.nan, np.nan, np.nan, np.nan],
#         [0.1, 0.1, 0.8, 0.15],
#         [0.2, 0.5, 0.3, -0.22],
#     ]))
